#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib
import json
import logging
import os
import sys
from ast import literal_eval
from datetime import datetime, timedelta
from time import sleep
from unittest import mock

# leave this it is used by the test worker
import celery.contrib.testing.tasks  # noqa: F401
import pytest
import uuid6
from celery import Celery
from celery.backends.base import BaseBackend, BaseKeyValueStoreBackend
from celery.backends.database import DatabaseBackend
from celery.contrib.testing.worker import start_worker
from kombu.asynchronous import set_event_loop
from kubernetes.client import models as k8s
from uuid6 import uuid7

from airflow._shared.timezones import timezone
from airflow.configuration import conf
from airflow.executors import workloads
from airflow.models.dag import DAG
from airflow.models.taskinstance import TaskInstance
from airflow.providers.common.compat.sdk import AirflowException, AirflowTaskTimeout, TaskInstanceKey
from airflow.providers.standard.operators.bash import BashOperator
from airflow.sdk import BaseOperator
from airflow.utils.state import State

from tests_common.test_utils import db
from tests_common.test_utils.taskinstance import create_task_instance
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

logger = logging.getLogger(__name__)


def _prepare_test_bodies():
    return literal_eval(os.environ["CELERY_BROKER_URLS_MAP"]).values()


class FakeCeleryResult:
    @property
    def state(self):
        raise Exception()

    def task_id(self):
        return "task_id"


@contextlib.contextmanager
def _prepare_app(broker_url=None, execute=None):
    from airflow.providers.celery.executors import celery_executor_utils

    broker_url = broker_url or conf.get("celery", "BROKER_URL")
    if AIRFLOW_V_3_0_PLUS:
        execute_name = "execute_workload"
        execute = execute or celery_executor_utils.execute_workload.__wrapped__
    else:
        execute_name = "execute_command"
        execute = execute or celery_executor_utils.execute_command.__wrapped__

    test_config = dict(celery_executor_utils.get_celery_configuration())
    test_config.update({"broker_url": broker_url})
    test_app = Celery(broker_url, config_source=test_config)
    # Register the fake execute function with the test_app using the correct task name
    # This ensures workers using test_app will execute the fake function
    test_execute = test_app.task(name=execute_name)(execute)
    patch_app = mock.patch.object(celery_executor_utils, "app", test_app)

    if AIRFLOW_V_3_0_PLUS:
        celery_executor_utils.execute_workload.__wrapped__ = execute
    else:
        celery_executor_utils.execute_command.__wrapped__ = execute

    patch_execute = mock.patch.object(celery_executor_utils, execute_name, test_execute)
    # Patch factory function so CeleryExecutor instances get the test app
    patch_factory = mock.patch.object(celery_executor_utils, "create_celery_app", return_value=test_app)

    backend = test_app.backend

    if hasattr(backend, "ResultSession"):
        # Pre-create the database tables now, otherwise SQLA vis Celery has a
        # race condition where it one of the subprocesses can die with "Table
        # already exists" error, because SQLA checks for which tables exist,
        # then issues a CREATE TABLE, rather than doing CREATE TABLE IF NOT
        # EXISTS
        session = backend.ResultSession()
        session.close()

    with patch_app, patch_execute, patch_factory:
        try:
            yield test_app
        finally:
            # Clear event loop to tear down each celery instance
            set_event_loop(None)


@pytest.mark.integration("celery")
@pytest.mark.backend("postgres")
class TestCeleryExecutor:
    def setup_method(self) -> None:
        db.clear_db_runs()
        db.clear_db_jobs()

    def teardown_method(self) -> None:
        db.clear_db_runs()
        db.clear_db_jobs()


def setup_dagrun_with_success_and_fail_tasks(dag_maker):
    date = timezone.utcnow()
    start_date = date - timedelta(days=2)

    with dag_maker("test_celery_integration"):
        BaseOperator(task_id="success", start_date=start_date)
        BaseOperator(task_id="fail", start_date=start_date)

    return dag_maker.create_dagrun(logical_date=date)

    @pytest.mark.flaky(reruns=5, reruns_delay=3)
    @pytest.mark.parametrize("broker_url", _prepare_test_bodies())
    @pytest.mark.parametrize(
        "executor_config",
        [
            pytest.param({}, id="no_executor_config"),
            pytest.param(
                {
                    "pod_override": k8s.V1Pod(
                        spec=k8s.V1PodSpec(
                            containers=[
                                k8s.V1Container(
                                    name="base",
                                    resources=k8s.V1ResourceRequirements(
                                        requests={
                                            "cpu": "100m",
                                            "memory": "384Mi",
                                        },
                                        limits={
                                            "cpu": 1,
                                            "memory": "500Mi",
                                        },
                                    ),
                                )
                            ]
                        )
                    )
                },
                id="pod_override_executor_config",
            ),
        ],
    )
    def test_celery_integration(self, broker_url, executor_config):
        from airflow.providers.celery.executors import celery_executor, celery_executor_utils

        if AIRFLOW_V_3_0_PLUS:
            # Airflow 3: execute_workload receives JSON string
            def fake_execute(input: str) -> None:
                """Fake execute_workload that parses JSON and fails for tasks with 'fail' in task_id."""
                import json

                workload_dict = json.loads(input)
                # Check if this is a task that should fail (task_id contains "fail")
                if "ti" in workload_dict and "task_id" in workload_dict["ti"]:
                    if "fail" in workload_dict["ti"]["task_id"]:
                        raise AirflowException("fail")
        else:
            # Airflow 2: execute_command receives command list
            def fake_execute(input: str) -> None:  # Use same parameter name as Airflow 3 version
                if "fail" in input:
                    raise AirflowException("fail")

        with _prepare_app(broker_url, execute=fake_execute) as app:
            executor = celery_executor.CeleryExecutor()
            assert executor.tasks == {}
            executor.start()

            with start_worker(app=app, logfile=sys.stdout, loglevel="info"):
                ti_success = workloads.TaskInstance.model_construct(
                    id=uuid7(),
                    task_id="success",
                    dag_id="id",
                    run_id="abc",
                    try_number=0,
                    priority_weight=1,
                    queue=celery_executor_utils.get_celery_configuration()["task_default_queue"],
                    executor_config=executor_config,
                )
                keys = [
                    TaskInstanceKey("id", "success", "abc", 0, -1),
                    TaskInstanceKey("id", "fail", "abc", 0, -1),
                ]
                dagrun = setup_dagrun_with_success_and_fail_tasks(dag_maker)
                ti_success, ti_fail = dagrun.task_instances
                for w in (
                    workloads.ExecuteTask.make(
                        ti=ti_success,
                    ),
                    workloads.ExecuteTask.make(ti=ti_fail),
                ):
                    executor.queue_workload(w, session=None)

                executor.trigger_tasks(open_slots=10)
                for _ in range(20):
                    num_tasks = len(executor.tasks.keys())
                    if num_tasks == 2:
                        break
                    logger.info(
                        "Waiting 0.1 s for tasks to be processed asynchronously. Processed so far %d",
                        num_tasks,
                    )
                    sleep(0.4)
                assert list(executor.tasks.keys()) == keys
                assert executor.event_buffer[keys[0]][0] == State.QUEUED
                assert executor.event_buffer[keys[1]][0] == State.QUEUED

                executor.end(synchronous=True)

        assert executor.event_buffer[keys[0]][0] == State.SUCCESS
        assert executor.event_buffer[keys[1]][0] == State.FAILED

        assert keys[0] not in executor.tasks
        assert keys[1] not in executor.tasks

        assert executor.queued_tasks == {}

    def test_error_sending_task(self):
        from airflow.providers.celery.executors import celery_executor, celery_executor_utils

        with _prepare_app():
            executor = celery_executor.CeleryExecutor()
            with DAG(dag_id="dag_id"):
                task = BashOperator(task_id="test", bash_command="true", start_date=datetime.now())
            if AIRFLOW_V_3_0_PLUS:
                ti = create_task_instance(task=task, run_id="abc", dag_version_id=uuid6.uuid7())
            else:
                ti = TaskInstance(task=task, run_id="abc")
            workload = workloads.ExecuteTask.model_construct(
                ti=workloads.TaskInstance.model_validate(ti, from_attributes=True),
            )

            key = (task.dag.dag_id, task.task_id, ti.run_id, 0, -1)
            executor.queued_tasks[key] = workload
            executor.task_publish_retries[key] = 1

            # Mock send_task_to_executor to return an error result
            # This simulates a failure when sending the task to Celery
            def mock_send_error(task_tuple):
                key_from_tuple = task_tuple[0]
                return (
                    key_from_tuple,
                    task_tuple[1],  # args
                    celery_executor_utils.ExceptionWithTraceback(
                        RuntimeError("Intentional test failure"),
                        "Celery Task ID: mock\nTraceback: test error",
                    ),
                )

            with mock.patch.object(
                celery_executor_utils, "send_task_to_executor", side_effect=mock_send_error
            ):
                executor.heartbeat()
        assert len(executor.queued_tasks) == 0, "Task should no longer be queued"
        assert executor.event_buffer[key][0] == State.FAILED

    def test_retry_on_error_sending_task(self, caplog):
        """Test that Airflow retries publishing tasks to Celery Broker at least 3 times"""
        from airflow.providers.celery.executors import celery_executor, celery_executor_utils

        with (
            _prepare_app(),
            caplog.at_level(logging.INFO),
            mock.patch.object(
                # Mock `with timeout()` to _instantly_ fail.
                celery_executor_utils.timeout,
                "__enter__",
                side_effect=AirflowTaskTimeout,
            ),
        ):
            executor = celery_executor.CeleryExecutor()
            assert executor.task_publish_retries == {}
            assert executor.task_publish_max_retries == 3, "Assert Default Max Retries is 3"

            with DAG(dag_id="id"):
                task = BashOperator(task_id="test", bash_command="true", start_date=datetime.now())
            if AIRFLOW_V_3_0_PLUS:
                ti = create_task_instance(task=task, run_id="abc", dag_version_id=uuid6.uuid7())
            else:
                ti = TaskInstance(task=task, run_id="abc")
            workload = workloads.ExecuteTask.model_construct(
                ti=workloads.TaskInstance.model_validate(ti, from_attributes=True),
            )

            key = (task.dag.dag_id, task.task_id, ti.run_id, 0, -1)
            executor.queued_tasks[key] = workload

            # Test that when heartbeat is called again, task is published again to Celery Queue
            executor.heartbeat()
            assert dict(executor.task_publish_retries) == {key: 1}
            assert len(executor.queued_tasks) == 1, "Task should remain in queue"
            assert executor.event_buffer == {}
            assert f"[Try 1 of 3] Task Timeout Error for Task: ({key})." in caplog.text

            executor.heartbeat()
            assert dict(executor.task_publish_retries) == {key: 2}
            assert len(executor.queued_tasks) == 1, "Task should remain in queue"
            assert executor.event_buffer == {}
            assert f"[Try 2 of 3] Task Timeout Error for Task: ({key})." in caplog.text

            executor.heartbeat()
            assert dict(executor.task_publish_retries) == {key: 3}
            assert len(executor.queued_tasks) == 1, "Task should remain in queue"
            assert executor.event_buffer == {}
            assert f"[Try 3 of 3] Task Timeout Error for Task: ({key})." in caplog.text

            executor.heartbeat()
            assert dict(executor.task_publish_retries) == {}
            assert len(executor.queued_tasks) == 0, "Task should no longer be in queue"
            assert executor.event_buffer[key][0] == State.FAILED


class ClassWithCustomAttributes:
    """Class for testing purpose: allows to create objects with custom attributes in one single statement."""

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __str__(self):
        return f"{ClassWithCustomAttributes.__name__}({str(self.__dict__)})"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __hash__(self):
        return hash(self.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)


@pytest.mark.integration("celery")
@pytest.mark.backend("postgres")
class TestBulkStateFetcher:
    bulk_state_fetcher_logger = "airflow.providers.celery.executors.celery_executor_utils.BulkStateFetcher"

    @mock.patch(
        "celery.backends.base.BaseKeyValueStoreBackend.mget",
        return_value=[json.dumps({"status": "SUCCESS", "task_id": "123"})],
    )
    def test_should_support_kv_backend(self, mock_mget, caplog):
        from airflow.providers.celery.executors import celery_executor, celery_executor_utils

        caplog.set_level(logging.DEBUG, logger=self.bulk_state_fetcher_logger)
        with _prepare_app():
            mock_backend = BaseKeyValueStoreBackend(app=celery_executor.app)
            with mock.patch(
                "airflow.providers.celery.executors.celery_executor_utils.Celery.backend", mock_backend
            ):
                caplog.clear()
                fetcher = celery_executor_utils.BulkStateFetcher(1)
                result = fetcher.get_many(
                    [
                        mock.MagicMock(task_id="123"),
                        mock.MagicMock(task_id="456"),
                    ]
                )

        # Assert called - ignore order
        mget_args, _ = mock_mget.call_args
        assert set(mget_args[0]) == {b"celery-task-meta-456", b"celery-task-meta-123"}
        mock_mget.assert_called_once_with(mock.ANY)

        assert result == {"123": ("SUCCESS", None), "456": ("PENDING", None)}
        assert caplog.messages == ["Fetched 2 state(s) for 2 task(s)"]

    @mock.patch("celery.backends.database.DatabaseBackend.ResultSession")
    def test_should_support_db_backend(self, mock_session, caplog):
        from airflow.providers.celery.executors import celery_executor, celery_executor_utils

        caplog.set_level(logging.DEBUG, logger=self.bulk_state_fetcher_logger)
        with _prepare_app():
            mock_backend = DatabaseBackend(app=celery_executor.app, url="sqlite3://")
            with mock.patch(
                "airflow.providers.celery.executors.celery_executor_utils.Celery.backend", mock_backend
            ):
                caplog.clear()
                mock_session = mock_backend.ResultSession.return_value
                mock_session.scalars.return_value.all.return_value = [
                    mock.MagicMock(**{"to_dict.return_value": {"status": "SUCCESS", "task_id": "123"}})
                ]

                fetcher = celery_executor_utils.BulkStateFetcher(1)
                result = fetcher.get_many(
                    [
                        mock.MagicMock(task_id="123"),
                        mock.MagicMock(task_id="456"),
                    ]
                )

        assert result == {"123": ("SUCCESS", None), "456": ("PENDING", None)}
        assert caplog.messages == ["Fetched 2 state(s) for 2 task(s)"]

    @mock.patch("celery.backends.database.DatabaseBackend.ResultSession")
    def test_should_retry_db_backend(self, mock_session, caplog):
        from airflow.providers.celery.executors import celery_executor, celery_executor_utils

        caplog.set_level(logging.DEBUG, logger=self.bulk_state_fetcher_logger)
        from sqlalchemy.exc import DatabaseError

        with _prepare_app():
            mock_backend = DatabaseBackend(app=celery_executor.app, url="sqlite3://")
            with mock.patch(
                "airflow.providers.celery.executors.celery_executor_utils.Celery.backend", mock_backend
            ):
                caplog.clear()
                mock_session = mock_backend.ResultSession.return_value
                mock_retry_db_result = mock_session.scalars.return_value.all
                mock_retry_db_result.return_value = [
                    mock.MagicMock(**{"to_dict.return_value": {"status": "SUCCESS", "task_id": "123"}})
                ]
                mock_retry_db_result.side_effect = [
                    DatabaseError("DatabaseError", "DatabaseError", "DatabaseError"),
                    mock_retry_db_result.return_value,
                ]

                fetcher = celery_executor_utils.BulkStateFetcher(1)
                result = fetcher.get_many(
                    [
                        mock.MagicMock(task_id="123"),
                        mock.MagicMock(task_id="456"),
                    ]
                )
        assert mock_retry_db_result.call_count == 2
        assert result == {"123": ("SUCCESS", None), "456": ("PENDING", None)}
        assert caplog.messages == [
            "Failed operation _query_task_cls_from_db_backend.  Retrying 2 more times.",
            "Fetched 2 state(s) for 2 task(s)",
        ]

    def test_should_support_base_backend(self, caplog):
        from airflow.providers.celery.executors import celery_executor_utils

        caplog.set_level(logging.DEBUG, logger=self.bulk_state_fetcher_logger)
        with _prepare_app():
            mock_backend = mock.MagicMock(autospec=BaseBackend)

            with mock.patch(
                "airflow.providers.celery.executors.celery_executor_utils.Celery.backend", mock_backend
            ):
                caplog.clear()
                fetcher = celery_executor_utils.BulkStateFetcher(1)
                result = fetcher.get_many(
                    [
                        ClassWithCustomAttributes(task_id="123", state="SUCCESS"),
                        ClassWithCustomAttributes(task_id="456", state="PENDING"),
                    ]
                )

        assert result == {"123": ("SUCCESS", None), "456": ("PENDING", None)}
        assert caplog.messages == ["Fetched 2 state(s) for 2 task(s)"]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Default celery configuration."""

from __future__ import annotations

import json
import logging
import re
import ssl
from typing import Any

from airflow.configuration import conf
from airflow.exceptions import AirflowConfigException
from airflow.providers.celery.version_compat import AIRFLOW_V_3_0_PLUS
from airflow.providers.common.compat.sdk import AirflowException

log = logging.getLogger(__name__)


def _broker_supports_visibility_timeout(url):
    return url.startswith(("redis://", "rediss://", "sqs://", "sentinel://"))


def get_default_celery_config(team_conf) -> dict[str, Any]:
    """
    Build Celery configuration using team-aware config.

    For Airflow versions < 3.2 that don't support multi-team configuration,
    falls back to using the global conf object.

    :param team_conf: ExecutorConf instance with team-specific configuration, or conf object
    :return: Dictionary with Celery configuration
    """
    # Check if team_conf supports team-specific config (has getsection method), if not then we know that
    # we're not running in a multi-team capable Airflow version, so we fallback to using the global conf.
    if not hasattr(team_conf, "getsection"):
        team_conf = conf

    broker_url = team_conf.get("celery", "BROKER_URL", fallback="redis://redis:6379/0")

    broker_transport_options: dict = team_conf.getsection("celery_broker_transport_options") or {}
    if "visibility_timeout" not in broker_transport_options:
        if _broker_supports_visibility_timeout(broker_url):
            broker_transport_options["visibility_timeout"] = 86400

    if "sentinel_kwargs" in broker_transport_options:
        try:
            sentinel_kwargs = json.loads(broker_transport_options["sentinel_kwargs"])
            if not isinstance(sentinel_kwargs, dict):
                raise ValueError
            broker_transport_options["sentinel_kwargs"] = sentinel_kwargs
        except Exception:
            raise AirflowException("sentinel_kwargs should be written in the correct dictionary format.")

    if team_conf.has_option("celery", "RESULT_BACKEND"):
        result_backend = team_conf.get_mandatory_value("celery", "RESULT_BACKEND")
    else:
        log.debug("Value for celery result_backend not found. Using sql_alchemy_conn with db+ prefix.")
        result_backend = f"db+{team_conf.get('database', 'SQL_ALCHEMY_CONN')}"

    # Handle result backend transport options (for Redis Sentinel support)
    result_backend_transport_options: dict = (
        team_conf.getsection("celery_result_backend_transport_options") or {}
    )
    if "sentinel_kwargs" in result_backend_transport_options:
        try:
            result_sentinel_kwargs = json.loads(result_backend_transport_options["sentinel_kwargs"])
            if not isinstance(result_sentinel_kwargs, dict):
                raise ValueError
            result_backend_transport_options["sentinel_kwargs"] = result_sentinel_kwargs
        except Exception:
            raise AirflowException(
                "sentinel_kwargs in [celery_result_backend_transport_options] should be written "
                "in the correct dictionary format."
            )

    extra_celery_config = team_conf.getjson("celery", "extra_celery_config", fallback={})

    config = {
        "accept_content": ["json"],
        "event_serializer": "json",
        "worker_prefetch_multiplier": team_conf.getint("celery", "worker_prefetch_multiplier", fallback=1),
        "task_acks_late": team_conf.getboolean("celery", "task_acks_late", fallback=True),
        "task_default_queue": team_conf.get("operators", "DEFAULT_QUEUE"),
        "task_default_exchange": team_conf.get("operators", "DEFAULT_QUEUE"),
        "task_track_started": team_conf.getboolean("celery", "task_track_started", fallback=True),
        "broker_url": broker_url,
        "broker_transport_options": broker_transport_options,
        "broker_connection_retry_on_startup": team_conf.getboolean(
            "celery", "broker_connection_retry_on_startup", fallback=True
        ),
        "result_backend": result_backend,
        "result_backend_transport_options": result_backend_transport_options,
        "database_engine_options": team_conf.getjson(
            "celery", "result_backend_sqlalchemy_engine_options", fallback={}
        ),
        "worker_concurrency": team_conf.getint("celery", "WORKER_CONCURRENCY", fallback=16),
        "worker_enable_remote_control": team_conf.getboolean(
            "celery", "worker_enable_remote_control", fallback=True
        ),
        **(extra_celery_config if isinstance(extra_celery_config, dict) else {}),
    }

    # In order to not change anything pre Task Execution API, we leave this setting as it was (unset) in Airflow2
    if AIRFLOW_V_3_0_PLUS:
        config.setdefault("worker_redirect_stdouts", False)
        config.setdefault("worker_hijack_root_logger", False)

    # Handle SSL configuration
    try:
        celery_ssl_active = team_conf.getboolean("celery", "SSL_ACTIVE", fallback=False)
    except AirflowConfigException:
        log.warning("Celery Executor will run without SSL")
        celery_ssl_active = False

    try:
        if celery_ssl_active:
            if broker_url and "amqp://" in broker_url:
                broker_use_ssl = {
                    "keyfile": team_conf.get("celery", "SSL_KEY"),
                    "certfile": team_conf.get("celery", "SSL_CERT"),
                    "ca_certs": team_conf.get("celery", "SSL_CACERT"),
                    "cert_reqs": ssl.CERT_REQUIRED,
                }
            elif broker_url and re.search("rediss?://|sentinel://", broker_url):
                broker_use_ssl = {
                    "ssl_keyfile": team_conf.get("celery", "SSL_KEY"),
                    "ssl_certfile": team_conf.get("celery", "SSL_CERT"),
                    "ssl_ca_certs": team_conf.get("celery", "SSL_CACERT"),
                    "ssl_cert_reqs": ssl.CERT_REQUIRED,
                }
            else:
                raise AirflowException(
                    "The broker you configured does not support SSL_ACTIVE to be True. "
                    "Please use RabbitMQ or Redis if you would like to use SSL for broker."
                )

            config["broker_use_ssl"] = broker_use_ssl
    except AirflowConfigException:
        raise AirflowException(
            "AirflowConfigException: SSL_ACTIVE is True, please ensure SSL_KEY, SSL_CERT and SSL_CACERT are set"
        )
    except Exception as e:
        raise AirflowException(
            f"Exception: There was an unknown Celery SSL Error. Please ensure you want to use SSL and/or have "
            f"all necessary certs and key ({e})."
        )

    # Warning for not recommended backends
    match_not_recommended_backend = re.search("rediss?://|amqp://|rpc://", result_backend)
    if match_not_recommended_backend:
        log.warning(
            "You have configured a result_backend using the protocol `%s`,"
            " it is highly recommended to use an alternative result_backend (i.e. a database).",
            match_not_recommended_backend.group(0).strip("://"),
        )

    return config


# IMPORTANT NOTE! Celery Executor has initialization done dynamically and it performs initialization when
# it is imported, so we need fallbacks here in order to be able to import the class directly without
# having configuration initialized before. Do not remove those fallbacks!
#
# This is not strictly needed for production:
#
#   * for Airflow 2.6 and before the defaults will come from the core defaults
#   * for Airflow 2.7+ the defaults will be loaded via ProvidersManager
#
# But it helps in our tests to import the executor class and validate if the celery code can be imported
# in the current and older versions of Airflow.

# For backward compatibility, keep DEFAULT_CELERY_CONFIG at module level using global config
# Use conf directly since we don't need team-specific config for the module-level default
DEFAULT_CELERY_CONFIG = get_default_celery_config(conf)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Utilities and classes used by the Celery Executor.

Much of this code is expensive to import/load, be careful where this module is imported.
"""

from __future__ import annotations

import contextlib
import logging
import math
import os
import subprocess
import sys
import traceback
from collections.abc import Collection, Mapping, MutableMapping, Sequence
from concurrent.futures import ProcessPoolExecutor
from functools import cache
from typing import TYPE_CHECKING, Any

from celery import Celery, states as celery_states
from celery.backends.base import BaseKeyValueStoreBackend
from celery.backends.database import DatabaseBackend, Task as TaskDb, retry, session_cleanup
from celery.signals import import_modules as celery_import_modules
from sqlalchemy import select

from airflow.configuration import AirflowConfigParser, conf
from airflow.executors.base_executor import BaseExecutor
from airflow.providers.celery.version_compat import AIRFLOW_V_3_0_PLUS
from airflow.providers.common.compat.sdk import AirflowException, AirflowTaskTimeout, Stats, timeout
from airflow.utils.log.logging_mixin import LoggingMixin
from airflow.utils.net import get_hostname
from airflow.utils.providers_configuration_loader import providers_configuration_loaded

try:
    from airflow.sdk.definitions._internal.dag_parsing_context import _airflow_parsing_context_manager
except ImportError:
    from airflow.utils.dag_parsing_context import _airflow_parsing_context_manager

log = logging.getLogger(__name__)

if sys.platform == "darwin":
    setproctitle = lambda title: log.debug("Mac OS detected, skipping setproctitle")
else:
    from setproctitle import setproctitle

if TYPE_CHECKING:
    from typing import TypeAlias

    from celery.result import AsyncResult

    from airflow.executors import workloads
    from airflow.executors.base_executor import EventBufferValueType, ExecutorConf
    from airflow.models.taskinstance import TaskInstanceKey

    # We can't use `if AIRFLOW_V_3_0_PLUS` conditions in type checks, so unfortunately we just have to define
    # the type as the union of both kinds
    CommandType = Sequence[str]

    TaskInstanceInCelery: TypeAlias = tuple[
        TaskInstanceKey, workloads.All | CommandType, str | None, str | None
    ]

    TaskTuple = tuple[TaskInstanceKey, CommandType, str | None, Any | None]

OPERATION_TIMEOUT = conf.getfloat("celery", "operation_timeout")

# Make it constant for unit test.
CELERY_FETCH_ERR_MSG_HEADER = "Error fetching Celery task state"


@cache
def get_celery_configuration() -> dict[str, Any]:
    """Get the Celery configuration dictionary."""
    if conf.has_option("celery", "celery_config_options"):
        return conf.getimport("celery", "celery_config_options")

    from airflow.providers.celery.executors.default_celery import DEFAULT_CELERY_CONFIG

    return DEFAULT_CELERY_CONFIG


@providers_configuration_loaded
def _get_celery_app() -> Celery:
    """Init providers before importing the configuration, so the _SECRET and _CMD options work."""
    celery_app_name = conf.get("celery", "CELERY_APP_NAME")

    return Celery(celery_app_name, config_source=get_celery_configuration())


def create_celery_app(team_conf: ExecutorConf | AirflowConfigParser) -> Celery:
    """
    Create a Celery app, supporting team-specific configuration.

    :param team_conf: ExecutorConf instance with team-specific configuration, or global conf
    :return: Celery app instance
    """
    from airflow.providers.celery.executors.default_celery import get_default_celery_config

    celery_app_name = team_conf.get("celery", "CELERY_APP_NAME")

    # Make app name unique per team to ensure proper broker isolation
    # Each team's executor needs a distinct Celery app name to prevent
    # tasks from being routed to the wrong broker
    # Only do this if team_conf is an ExecutorConf with team_name (not global conf)
    team_name = getattr(team_conf, "team_name", None)
    if team_name:
        celery_app_name = f"{celery_app_name}_{team_name}"

    config = get_default_celery_config(team_conf)

    celery_app = Celery(celery_app_name, config_source=config)

    # Register tasks with this app
    celery_app.task(name="execute_workload")(execute_workload)
    if not AIRFLOW_V_3_0_PLUS:
        celery_app.task(name="execute_command")(execute_command)

    return celery_app


# Keep module-level app for backward compatibility
app = _get_celery_app()


@celery_import_modules.connect
def on_celery_import_modules(*args, **kwargs):
    """
    Preload some "expensive" airflow modules once, so other task processes won't have to import it again.

    Loading these for each task adds 0.3-0.5s *per task* before the task can run. For long-running tasks this
    doesn't matter, but for short tasks this starts to be a noticeable impact.
    """
    import jinja2.ext  # noqa: F401

    if not AIRFLOW_V_3_0_PLUS:
        import airflow.jobs.local_task_job_runner
        import airflow.macros

    try:
        import airflow.providers.standard.operators.bash
        import airflow.providers.standard.operators.python
    except ImportError:
        import airflow.operators.bash
        import airflow.operators.python  # noqa: F401

    with contextlib.suppress(ImportError):
        import numpy  # noqa: F401

    with contextlib.suppress(ImportError):
        import kubernetes.client  # noqa: F401


# Once Celery 5.5 is out of beta, we can pass `pydantic=True` to the decorator and it will handle the validation
# and deserialization for us
@app.task(name="execute_workload")
def execute_workload(input: str) -> None:
    from pydantic import TypeAdapter

    from airflow.configuration import conf
    from airflow.executors import workloads
    from airflow.sdk.execution_time.supervisor import supervise

    decoder = TypeAdapter[workloads.All](workloads.All)
    workload = decoder.validate_json(input)

    celery_task_id = app.current_task.request.id

    if not isinstance(workload, workloads.ExecuteTask):
        raise ValueError(f"CeleryExecutor does not know how to handle {type(workload)}")

    log.info("[%s] Executing workload in Celery: %s", celery_task_id, workload)

    base_url = conf.get("api", "base_url", fallback="/")
    # If it's a relative URL, use localhost:8080 as the default
    if base_url.startswith("/"):
        base_url = f"http://localhost:8080{base_url}"
    default_execution_api_server = f"{base_url.rstrip('/')}/execution/"

    supervise(
        # This is the "wrong" ti type, but it duck types the same. TODO: Create a protocol for this.
        ti=workload.ti,  # type: ignore[arg-type]
        dag_rel_path=workload.dag_rel_path,
        bundle_info=workload.bundle_info,
        token=workload.token,
        server=conf.get("core", "execution_api_server_url", fallback=default_execution_api_server),
        log_path=workload.log_path,
    )


if not AIRFLOW_V_3_0_PLUS:

    @app.task
    def execute_command(command_to_exec: CommandType) -> None:
        """Execute command."""
        EXECUTE_TASKS_NEW_PYTHON_INTERPRETER = not hasattr(os, "fork") or conf.getboolean(
            "core",
            "execute_tasks_new_python_interpreter",
            fallback=False,
        )

        dag_id, task_id = BaseExecutor.validate_airflow_tasks_run_command(command_to_exec)  # type: ignore[attr-defined]
        celery_task_id = app.current_task.request.id
        log.info("[%s] Executing command in Celery: %s", celery_task_id, command_to_exec)
        with _airflow_parsing_context_manager(dag_id=dag_id, task_id=task_id):
            try:
                if EXECUTE_TASKS_NEW_PYTHON_INTERPRETER:
                    _execute_in_subprocess(command_to_exec, celery_task_id)
                else:
                    _execute_in_fork(command_to_exec, celery_task_id)
            except Exception:
                Stats.incr("celery.execute_command.failure")
                raise


def _execute_in_fork(command_to_exec: CommandType, celery_task_id: str | None = None) -> None:
    pid = os.fork()
    if pid:
        # In parent, wait for the child
        pid, ret = os.waitpid(pid, 0)
        if ret == 0:
            return

        msg = f"Celery command failed on host: {get_hostname()} with celery_task_id {celery_task_id} (PID: {pid}, Return Code: {ret})"
        raise AirflowException(msg)

    from airflow.sentry import Sentry

    ret = 1
    try:
        from airflow.cli.cli_parser import get_parser

        parser = get_parser()
        # [1:] - remove "airflow" from the start of the command
        args = parser.parse_args(command_to_exec[1:])
        args.shut_down_logging = False
        if celery_task_id:
            args.external_executor_id = celery_task_id

        setproctitle(f"airflow task supervisor: {command_to_exec}")
        log.debug("calling func '%s' with args %s", args.func.__name__, args)
        args.func(args)
        ret = 0
    except Exception:
        log.exception("[%s] Failed to execute task.", celery_task_id)
        ret = 1
    finally:
        try:
            Sentry.flush()
            logging.shutdown()
        except Exception:
            log.exception("[%s] Failed to clean up.", celery_task_id)
            ret = 1
        os._exit(ret)


def _execute_in_subprocess(command_to_exec: CommandType, celery_task_id: str | None = None) -> None:
    env = os.environ.copy()
    if celery_task_id:
        env["external_executor_id"] = celery_task_id
    try:
        subprocess.run(
            command_to_exec,
            check=False,
            stderr=sys.__stderr__,
            stdout=sys.__stdout__,
            close_fds=True,
            env=env,
        )
    except subprocess.CalledProcessError as e:
        log.exception("[%s] execute_command encountered a CalledProcessError", celery_task_id)
        log.error(e.output)
        msg = f"Celery command failed on host: {get_hostname()} with celery_task_id {celery_task_id}"
        raise AirflowException(msg)


class ExceptionWithTraceback:
    """
    Wrapper class used to propagate exceptions to parent processes from subprocesses.

    :param exception: The exception to wrap
    :param exception_traceback: The stacktrace to wrap
    """

    def __init__(self, exception: BaseException, exception_traceback: str):
        self.exception = exception
        self.traceback = exception_traceback


def send_task_to_executor(
    task_tuple: TaskInstanceInCelery,
) -> tuple[TaskInstanceKey, CommandType, AsyncResult | ExceptionWithTraceback]:
    """
    Send task to executor.

    This function is called in ProcessPoolExecutor subprocesses. To avoid pickling issues with
    team-specific Celery apps, we pass the team_name and reconstruct the Celery app here.
    """
    key, args, queue, team_name = task_tuple

    # Reconstruct the Celery app from configuration, which may or may not be team-specific.
    # ExecutorConf wraps config access to automatically use team-specific config where present.
    if TYPE_CHECKING:
        _conf: ExecutorConf | AirflowConfigParser
    # Check if Airflow version is greater than or equal to 3.2 to import ExecutorConf
    if AIRFLOW_V_3_0_PLUS:
        from airflow.executors.base_executor import ExecutorConf

        _conf = ExecutorConf(team_name)
    else:
        # Airflow <3.2 ExecutorConf doesn't exist (at least not with the required attributes), fall back to global conf
        _conf = conf

    # Create the Celery app with the correct configuration
    celery_app = create_celery_app(_conf)

    if AIRFLOW_V_3_0_PLUS:
        # Get the task from the app
        task_to_run = celery_app.tasks["execute_workload"]
        if TYPE_CHECKING:
            assert isinstance(args, workloads.BaseWorkload)
        args = (args.model_dump_json(),)
    else:
        # Get the task from the app
        task_to_run = celery_app.tasks["execute_command"]
        args = [args]  # type: ignore[list-item]

    # Pre-import redis.client to avoid SIGALRM interrupting module initialization.
    # If timeout fires during import, redis module gets partially cached in sys.modules
    # without the 'client' submodule bound, causing AttributeError on subsequent access.
    # See: https://github.com/apache/airflow/issues/41359
    try:
        import redis.client  # noqa: F401
    except ImportError:
        pass  # Redis not installed or not using Redis backend

    try:
        with timeout(seconds=OPERATION_TIMEOUT):
            result = task_to_run.apply_async(args=args, queue=queue)
    except (Exception, AirflowTaskTimeout) as e:
        exception_traceback = f"Celery Task ID: {key}\n{traceback.format_exc()}"
        result = ExceptionWithTraceback(e, exception_traceback)

    # The type is right for the version, but the type cannot be defined correctly for Airflow 2 and 3
    # concurrently;
    return key, args, result


def fetch_celery_task_state(async_result: AsyncResult) -> tuple[str, str | ExceptionWithTraceback, Any]:
    """
    Fetch and return the state of the given celery task.

    The scope of this function is global so that it can be called by subprocesses in the pool.

    :param async_result: a tuple of the Celery task key and the async Celery object used
        to fetch the task's state
    :return: a tuple of the Celery task key and the Celery state and the celery info
        of the task
    """
    # Pre-import redis.client to avoid SIGALRM interrupting module initialization.
    # See: https://github.com/apache/airflow/issues/41359
    try:
        import redis.client  # noqa: F401
    except ImportError:
        pass  # Redis not installed or not using Redis backend

    try:
        with timeout(seconds=OPERATION_TIMEOUT):
            # Accessing state property of celery task will make actual network request
            # to get the current state of the task
            info = async_result.info if hasattr(async_result, "info") else None
            return async_result.task_id, async_result.state, info
    except Exception as e:
        exception_traceback = f"Celery Task ID: {async_result}\n{traceback.format_exc()}"
        return async_result.task_id, ExceptionWithTraceback(e, exception_traceback), None


class BulkStateFetcher(LoggingMixin):
    """
    Gets status for many Celery tasks using the best method available.

    If BaseKeyValueStoreBackend is used as result backend, the mget method is used.
    If DatabaseBackend is used as result backend, the SELECT ...WHERE task_id IN (...) query is used
    Otherwise, multiprocessing.Pool will be used. Each task status will be downloaded individually.
    """

    def __init__(self, sync_parallelism: int, celery_app: Celery | None = None):
        super().__init__()
        self._sync_parallelism = sync_parallelism
        self.celery_app = celery_app or app  # Use provided app or fall back to module-level app

    def _tasks_list_to_task_ids(self, async_tasks: Collection[AsyncResult]) -> set[str]:
        return {a.task_id for a in async_tasks}

    def get_many(self, async_results: Collection[AsyncResult]) -> Mapping[str, EventBufferValueType]:
        """Get status for many Celery tasks using the best method available."""
        if isinstance(self.celery_app.backend, BaseKeyValueStoreBackend):
            result = self._get_many_from_kv_backend(async_results)
        elif isinstance(self.celery_app.backend, DatabaseBackend):
            result = self._get_many_from_db_backend(async_results)
        else:
            result = self._get_many_using_multiprocessing(async_results)
        self.log.debug("Fetched %d state(s) for %d task(s)", len(result), len(async_results))
        return result

    def _get_many_from_kv_backend(
        self, async_tasks: Collection[AsyncResult]
    ) -> Mapping[str, EventBufferValueType]:
        task_ids = self._tasks_list_to_task_ids(async_tasks)
        keys = [self.celery_app.backend.get_key_for_task(k) for k in task_ids]
        values = self.celery_app.backend.mget(keys)
        task_results = [self.celery_app.backend.decode_result(v) for v in values if v]
        task_results_by_task_id = {task_result["task_id"]: task_result for task_result in task_results}

        return self._prepare_state_and_info_by_task_dict(task_ids, task_results_by_task_id)

    @retry
    def _query_task_cls_from_db_backend(self, task_ids: set[str], **kwargs):
        session = self.celery_app.backend.ResultSession()
        task_cls = getattr(self.celery_app.backend, "task_cls", TaskDb)
        with session_cleanup(session):
            return session.scalars(select(task_cls).where(task_cls.task_id.in_(task_ids))).all()

    def _get_many_from_db_backend(
        self, async_tasks: Collection[AsyncResult]
    ) -> Mapping[str, EventBufferValueType]:
        task_ids = self._tasks_list_to_task_ids(async_tasks)
        tasks = self._query_task_cls_from_db_backend(task_ids)
        task_results = [self.celery_app.backend.meta_from_decoded(task.to_dict()) for task in tasks]
        task_results_by_task_id = {task_result["task_id"]: task_result for task_result in task_results}

        return self._prepare_state_and_info_by_task_dict(task_ids, task_results_by_task_id)

    @staticmethod
    def _prepare_state_and_info_by_task_dict(
        task_ids: set[str], task_results_by_task_id: dict[str, dict[str, Any]]
    ) -> Mapping[str, EventBufferValueType]:
        state_info: MutableMapping[str, EventBufferValueType] = {}
        for task_id in task_ids:
            task_result = task_results_by_task_id.get(task_id)
            if task_result:
                state = task_result["status"]
                info = task_result.get("info")
            else:
                state = celery_states.PENDING
                info = None
            state_info[task_id] = state, info
        return state_info

    def _get_many_using_multiprocessing(
        self, async_results: Collection[AsyncResult]
    ) -> Mapping[str, EventBufferValueType]:
        num_process = min(len(async_results), self._sync_parallelism)

        with ProcessPoolExecutor(max_workers=num_process) as sync_pool:
            chunksize = max(1, math.ceil(len(async_results) / self._sync_parallelism))

            task_id_to_states_and_info = list(
                sync_pool.map(fetch_celery_task_state, async_results, chunksize=chunksize)
            )

            states_and_info_by_task_id: MutableMapping[str, EventBufferValueType] = {}
            for task_id, state_or_exception, info in task_id_to_states_and_info:
                if isinstance(state_or_exception, ExceptionWithTraceback):
                    self.log.error(
                        "%s:%s\n%s\n",
                        CELERY_FETCH_ERR_MSG_HEADER,
                        state_or_exception.exception,
                        state_or_exception.traceback,
                    )
                else:
                    states_and_info_by_task_id[task_id] = state_or_exception, info
        return states_and_info_by_task_id



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib
import logging
import os
import signal
import sys
from datetime import timedelta
from unittest import mock

# leave this it is used by the test worker
import celery.contrib.testing.tasks  # noqa: F401
import pytest
import time_machine
from celery import Celery
from celery.result import AsyncResult
from kombu.asynchronous import set_event_loop

from airflow.configuration import conf
from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.models.dag import DAG
from airflow.models.taskinstance import TaskInstance, TaskInstanceKey
from airflow.providers.celery.executors import celery_executor, celery_executor_utils, default_celery
from airflow.providers.celery.executors.celery_executor import CeleryExecutor
from airflow.utils.state import State

from tests_common.test_utils import db
from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.dag import sync_dag_to_db
from tests_common.test_utils.taskinstance import create_task_instance
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS, AIRFLOW_V_3_1_PLUS, AIRFLOW_V_3_2_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.models.dag_version import DagVersion
if AIRFLOW_V_3_1_PLUS:
    from airflow.sdk import BaseOperator, timezone
else:
    from airflow.models.baseoperator import BaseOperator  # type: ignore[attr-defined,no-redef]
    from airflow.utils import timezone  # type: ignore[attr-defined,no-redef]

pytestmark = pytest.mark.db_test


FAKE_EXCEPTION_MSG = "Fake Exception"


def _prepare_test_bodies():
    if "CELERY_BROKER_URLS" in os.environ:
        return [(url,) for url in os.environ["CELERY_BROKER_URLS"].split(",")]
    return [(conf.get("celery", "BROKER_URL"))]


class FakeCeleryResult:
    @property
    def state(self):
        raise Exception(FAKE_EXCEPTION_MSG)

    def task_id(self):
        return "task_id"


@contextlib.contextmanager
def _prepare_app(broker_url=None, execute=None):
    broker_url = broker_url or conf.get("celery", "BROKER_URL")

    if AIRFLOW_V_3_0_PLUS:
        execute_name = "execute_workload"
        execute = execute or celery_executor_utils.execute_workload.__wrapped__
    else:
        execute_name = "execute_command"
        execute = execute or celery_executor_utils.execute_command.__wrapped__

    test_config = dict(celery_executor_utils.get_celery_configuration())
    test_config.update({"broker_url": broker_url})
    test_app = Celery(broker_url, config_source=test_config)
    test_execute = test_app.task(execute)
    patch_app = mock.patch.object(celery_executor_utils, "app", test_app)
    patch_execute = mock.patch.object(celery_executor_utils, execute_name, test_execute)
    # Patch factory function so CeleryExecutor instances get the test app
    patch_factory = mock.patch.object(celery_executor_utils, "create_celery_app", return_value=test_app)

    backend = test_app.backend

    if hasattr(backend, "ResultSession"):
        # Pre-create the database tables now, otherwise SQLA via Celery has a
        # race condition where it one of the subprocesses can die with "Table
        # already exists" error, because SQLA checks for which tables exist,
        # then issues a CREATE TABLE, rather than doing CREATE TABLE IF NOT
        # EXISTS
        session = backend.ResultSession()
        session.close()

    with patch_app, patch_execute, patch_factory:
        try:
            yield test_app
        finally:
            # Clear event loop to tear down each celery instance
            set_event_loop(None)


class TestCeleryExecutor:
    def setup_method(self) -> None:
        db.clear_db_runs()
        db.clear_db_jobs()

    def teardown_method(self) -> None:
        db.clear_db_runs()
        db.clear_db_jobs()

    @pytest.mark.skipif(not AIRFLOW_V_3_2_PLUS, reason="Airflow 3.2+ prefers new configuration")
    def test_sentry_integration(self):
        assert CeleryExecutor.sentry_integration == "sentry_sdk.integrations.celery.CeleryIntegration"

    @pytest.mark.skipif(AIRFLOW_V_3_2_PLUS, reason="Test only for Airflow < 3.2")
    def test_supports_sentry(self):
        assert CeleryExecutor.supports_sentry

    def test_cli_commands_vended(self):
        assert CeleryExecutor.get_cli_commands()

    def test_celery_executor_init_with_args_kwargs(self):
        """Test that CeleryExecutor properly passes args and kwargs to BaseExecutor."""
        parallelism = 50
        team_name = "test_team"

        if AIRFLOW_V_3_2_PLUS:
            # Multi-team support with ExecutorConf requires Airflow 3.2+
            executor = celery_executor.CeleryExecutor(parallelism=parallelism, team_name=team_name)
        else:
            executor = celery_executor.CeleryExecutor(parallelism)

        assert executor.parallelism == parallelism

        if AIRFLOW_V_3_2_PLUS:
            # Multi-team support with ExecutorConf requires Airflow 3.2+
            assert executor.team_name == team_name
            assert executor.conf.team_name == team_name

    @pytest.mark.backend("mysql", "postgres")
    def test_exception_propagation(self, caplog):
        caplog.set_level(
            logging.ERROR, logger="airflow.providers.celery.executors.celery_executor_utils.BulkStateFetcher"
        )
        with _prepare_app():
            executor = celery_executor.CeleryExecutor()
            executor.tasks = {"key": FakeCeleryResult()}
            executor.bulk_state_fetcher._get_many_using_multiprocessing(executor.tasks.values())
        assert celery_executor_utils.CELERY_FETCH_ERR_MSG_HEADER in caplog.text, caplog.record_tuples
        assert FAKE_EXCEPTION_MSG in caplog.text, caplog.record_tuples

    @mock.patch("airflow.providers.celery.executors.celery_executor.CeleryExecutor.sync")
    @mock.patch("airflow.providers.celery.executors.celery_executor.CeleryExecutor.trigger_tasks")
    @mock.patch("airflow.executors.base_executor.Stats.gauge")
    def test_gauge_executor_metrics(self, mock_stats_gauge, mock_trigger_tasks, mock_sync):
        executor = celery_executor.CeleryExecutor()
        executor.heartbeat()
        calls = [
            mock.call(
                "executor.open_slots", value=mock.ANY, tags={"status": "open", "name": "CeleryExecutor"}
            ),
            mock.call(
                "executor.queued_tasks", value=mock.ANY, tags={"status": "queued", "name": "CeleryExecutor"}
            ),
            mock.call(
                "executor.running_tasks", value=mock.ANY, tags={"status": "running", "name": "CeleryExecutor"}
            ),
        ]
        mock_stats_gauge.assert_has_calls(calls)

    @pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="Airflow 3 doesn't have execute_command anymore")
    @pytest.mark.parametrize(
        ("command", "raise_exception"),
        [
            pytest.param(["true"], True, id="wrong-command"),
            pytest.param(["airflow", "tasks"], True, id="incomplete-command"),
            pytest.param(["airflow", "tasks", "run"], False, id="complete-command"),
        ],
    )
    def test_command_validation(self, command, raise_exception):
        """Check that we validate _on the receiving_ side, not just sending side"""
        expected_context = contextlib.nullcontext()
        if raise_exception:
            expected_context = pytest.raises(
                ValueError, match=r'The command must start with \["airflow", "tasks", "run"\]\.'
            )

        with (
            mock.patch(
                "airflow.providers.celery.executors.celery_executor_utils._execute_in_subprocess"
            ) as mock_subproc,
            mock.patch(
                "airflow.providers.celery.executors.celery_executor_utils._execute_in_fork"
            ) as mock_fork,
            mock.patch("celery.app.task.Task.request") as mock_task,
        ):
            mock_task.id = "abcdef-124215-abcdef"
            with expected_context:
                celery_executor_utils.execute_command(command)
            if raise_exception:
                mock_subproc.assert_not_called()
                mock_fork.assert_not_called()
            else:
                # One of these should be called.
                assert mock_subproc.call_args == (
                    (command, "abcdef-124215-abcdef"),
                ) or mock_fork.call_args == ((command, "abcdef-124215-abcdef"),)

    @pytest.mark.backend("mysql", "postgres")
    def test_try_adopt_task_instances_none(self, clean_dags_dagruns_and_dagbundles, testing_dag_bundle):
        start_date = timezone.utcnow() - timedelta(days=2)

        with DAG("test_try_adopt_task_instances_none", schedule=None) as dag:
            task_1 = BaseOperator(task_id="task_1", start_date=start_date)

        if AIRFLOW_V_3_0_PLUS:
            sync_dag_to_db(dag)
            dag_version = DagVersion.get_latest_version(dag.dag_id)
            key1 = create_task_instance(task=task_1, run_id=None, dag_version_id=dag_version.id)
        else:
            key1 = TaskInstance(task=task_1, run_id=None)
        tis = [key1]

        executor = celery_executor.CeleryExecutor()

        assert executor.try_adopt_task_instances(tis) == tis

    @pytest.mark.backend("mysql", "postgres")
    @time_machine.travel("2020-01-01", tick=False)
    def test_try_adopt_task_instances(self, clean_dags_dagruns_and_dagbundles, testing_dag_bundle):
        start_date = timezone.utcnow() - timedelta(days=2)

        with DAG("test_try_adopt_task_instances_none", schedule=None) as dag:
            task_1 = BaseOperator(task_id="task_1", start_date=start_date)
            task_2 = BaseOperator(task_id="task_2", start_date=start_date)

        if AIRFLOW_V_3_0_PLUS:
            sync_dag_to_db(dag)
            dag_version = DagVersion.get_latest_version(dag.dag_id)
            ti1 = create_task_instance(task=task_1, run_id=None, dag_version_id=dag_version.id)
            ti2 = create_task_instance(task=task_2, run_id=None, dag_version_id=dag_version.id)
        else:
            ti1 = TaskInstance(task=task_1, run_id=None)
            ti2 = TaskInstance(task=task_2, run_id=None)
        ti1.external_executor_id = "231"
        ti1.state = State.QUEUED
        ti2.external_executor_id = "232"
        ti2.state = State.QUEUED

        tis = [ti1, ti2]

        executor = celery_executor.CeleryExecutor()
        assert executor.running == set()
        assert executor.tasks == {}

        not_adopted_tis = executor.try_adopt_task_instances(tis)

        key_1 = TaskInstanceKey(dag.dag_id, task_1.task_id, None, 0)
        key_2 = TaskInstanceKey(dag.dag_id, task_2.task_id, None, 0)
        assert executor.running == {key_1, key_2}

        assert executor.tasks == {key_1: AsyncResult("231"), key_2: AsyncResult("232")}
        assert not_adopted_tis == []

    @pytest.fixture
    def mock_celery_revoke(self):
        with _prepare_app() as app:
            app.control.revoke = mock.MagicMock()
            yield app.control.revoke

    @pytest.mark.backend("mysql", "postgres")
    @mock.patch("airflow.providers.celery.executors.celery_executor.CeleryExecutor.fail")
    def test_cleanup_stuck_queued_tasks(
        self, mock_fail, clean_dags_dagruns_and_dagbundles, testing_dag_bundle
    ):
        start_date = timezone.utcnow() - timedelta(days=2)

        with DAG("test_cleanup_stuck_queued_tasks_failed", schedule=None) as dag:
            task = BaseOperator(task_id="task_1", start_date=start_date)

        if AIRFLOW_V_3_0_PLUS:
            sync_dag_to_db(dag)
            dag_version = DagVersion.get_latest_version(task.dag.dag_id)
            ti = create_task_instance(task=task, run_id=None, dag_version_id=dag_version.id)
        else:
            ti = TaskInstance(task=task, run_id=None)
        ti.external_executor_id = "231"
        ti.state = State.QUEUED
        ti.queued_dttm = timezone.utcnow() - timedelta(minutes=30)
        ti.queued_by_job_id = 1
        tis = [ti]
        with _prepare_app() as app:
            app.control.revoke = mock.MagicMock()
            executor = celery_executor.CeleryExecutor()
            executor.job_id = 1
            executor.running = {ti.key}
            executor.tasks = {ti.key: AsyncResult("231")}
            assert executor.has_task(ti)
            with pytest.warns(AirflowProviderDeprecationWarning, match="cleanup_stuck_queued_tasks"):
                executor.cleanup_stuck_queued_tasks(tis=tis)
            executor.sync()
        assert executor.tasks == {}
        app.control.revoke.assert_called_once_with("231")
        mock_fail.assert_called()
        assert not executor.has_task(ti)

    @pytest.mark.backend("mysql", "postgres")
    @mock.patch("airflow.providers.celery.executors.celery_executor.CeleryExecutor.fail")
    def test_revoke_task(self, mock_fail, clean_dags_dagruns_and_dagbundles, testing_dag_bundle):
        start_date = timezone.utcnow() - timedelta(days=2)

        with DAG("test_revoke_task", schedule=None) as dag:
            task = BaseOperator(task_id="task_1", start_date=start_date)

        if AIRFLOW_V_3_0_PLUS:
            sync_dag_to_db(dag)
            dag_version = DagVersion.get_latest_version(task.dag.dag_id)
            ti = create_task_instance(task=task, run_id=None, dag_version_id=dag_version.id)
        else:
            ti = TaskInstance(task=task, run_id=None)
        ti.external_executor_id = "231"
        ti.state = State.QUEUED
        ti.queued_dttm = timezone.utcnow() - timedelta(minutes=30)
        ti.queued_by_job_id = 1
        tis = [ti]
        with _prepare_app() as app:
            app.control.revoke = mock.MagicMock()
            executor = celery_executor.CeleryExecutor()
            executor.job_id = 1
            executor.running = {ti.key}
            executor.tasks = {ti.key: AsyncResult("231")}
            assert executor.has_task(ti)
            for ti in tis:
                executor.revoke_task(ti=ti)
            executor.sync()
        app.control.revoke.assert_called_once_with("231")
        assert executor.tasks == {}
        assert not executor.has_task(ti)
        mock_fail.assert_not_called()

    @conf_vars({("celery", "result_backend_sqlalchemy_engine_options"): '{"pool_recycle": 1800}'})
    def test_result_backend_sqlalchemy_engine_options(self):
        import importlib

        # Scope the mock using context manager so we can clean up afterward
        with mock.patch("celery.Celery") as mock_celery:
            # reload celery conf to apply the new config
            importlib.reload(default_celery)
            # reload celery_executor_utils to recreate the celery app with new config
            importlib.reload(celery_executor_utils)

            call_args = mock_celery.call_args.kwargs.get("config_source")
            assert "database_engine_options" in call_args
            assert call_args["database_engine_options"] == {"pool_recycle": 1800}

        # Clean up: reload modules with real Celery to restore clean state for subsequent tests
        importlib.reload(default_celery)
        importlib.reload(celery_executor_utils)


def test_operation_timeout_config():
    assert celery_executor_utils.OPERATION_TIMEOUT == 1


class MockTask:
    """
    A picklable object used to mock tasks sent to Celery. Can't use the mock library
    here because it's not picklable.
    """

    def apply_async(self, *args, **kwargs):
        return 1


def _exit_gracefully(signum, _):
    print(f"{os.getpid()} Exiting gracefully upon receiving signal {signum}")
    sys.exit(signum)


@pytest.fixture
def register_signals():
    """
    Register the same signals as scheduler does to test celery_executor to make sure it does not
    hang.
    """
    orig_sigint = orig_sigterm = orig_sigusr2 = signal.SIG_DFL

    orig_sigint = signal.signal(signal.SIGINT, _exit_gracefully)
    orig_sigterm = signal.signal(signal.SIGTERM, _exit_gracefully)
    orig_sigusr2 = signal.signal(signal.SIGUSR2, _exit_gracefully)

    yield

    # Restore original signal handlers after test
    signal.signal(signal.SIGINT, orig_sigint)
    signal.signal(signal.SIGTERM, orig_sigterm)
    signal.signal(signal.SIGUSR2, orig_sigusr2)


@pytest.mark.execution_timeout(200)
@pytest.mark.quarantined
def test_send_tasks_to_celery_hang(register_signals):
    """
    Test that celery_executor does not hang after many runs.
    """
    executor = celery_executor.CeleryExecutor()

    task = MockTask()
    task_tuples_to_send = [(None, None, None, task) for _ in range(26)]

    for _ in range(250):
        # This loop can hang on Linux if celery_executor does something wrong with
        # multiprocessing.
        results = executor._send_tasks_to_celery(task_tuples_to_send)
        assert results == [(None, None, 1) for _ in task_tuples_to_send]


@conf_vars({("celery", "result_backend"): "rediss://test_user:test_password@localhost:6379/0"})
def test_celery_executor_with_no_recommended_result_backend(caplog):
    import importlib

    from airflow.providers.celery.executors.default_celery import log

    with caplog.at_level(logging.WARNING, logger=log.name):
        # reload celery conf to apply the new config
        importlib.reload(default_celery)
        assert "test_password" not in caplog.text
        assert (
            "You have configured a result_backend using the protocol `rediss`,"
            " it is highly recommended to use an alternative result_backend (i.e. a database)."
        ) in caplog.text


@conf_vars({("celery_broker_transport_options", "sentinel_kwargs"): '{"service_name": "mymaster"}'})
def test_sentinel_kwargs_loaded_from_string():
    import importlib

    # reload celery conf to apply the new config
    importlib.reload(default_celery)
    assert default_celery.DEFAULT_CELERY_CONFIG["broker_transport_options"]["sentinel_kwargs"] == {
        "service_name": "mymaster"
    }


@conf_vars({("celery", "task_acks_late"): "False"})
def test_celery_task_acks_late_loaded_from_string():
    import importlib

    # reload celery conf to apply the new config
    importlib.reload(default_celery)
    assert default_celery.DEFAULT_CELERY_CONFIG["task_acks_late"] is False


@conf_vars({("celery", "extra_celery_config"): '{"worker_max_tasks_per_child": 10}'})
def test_celery_extra_celery_config_loaded_from_string():
    import importlib

    # reload celery conf to apply the new config
    importlib.reload(default_celery)
    assert default_celery.DEFAULT_CELERY_CONFIG["worker_max_tasks_per_child"] == 10


@conf_vars({("celery_result_backend_transport_options", "sentinel_kwargs"): '{"password": "redis_password"}'})
def test_result_backend_sentinel_kwargs_loaded_from_string():
    """Test that sentinel_kwargs for result backend transport options is correctly parsed."""
    import importlib

    # reload celery conf to apply the new config
    importlib.reload(default_celery)
    assert "result_backend_transport_options" in default_celery.DEFAULT_CELERY_CONFIG
    assert default_celery.DEFAULT_CELERY_CONFIG["result_backend_transport_options"]["sentinel_kwargs"] == {
        "password": "redis_password"
    }


@conf_vars({("celery_result_backend_transport_options", "master_name"): "mymaster"})
def test_result_backend_master_name_loaded():
    """Test that master_name for result backend transport options is correctly loaded."""
    import importlib

    # reload celery conf to apply the new config
    importlib.reload(default_celery)
    assert "result_backend_transport_options" in default_celery.DEFAULT_CELERY_CONFIG
    assert (
        default_celery.DEFAULT_CELERY_CONFIG["result_backend_transport_options"]["master_name"] == "mymaster"
    )


@conf_vars(
    {
        ("celery_result_backend_transport_options", "sentinel_kwargs"): '{"password": "redis_password"}',
        ("celery_result_backend_transport_options", "master_name"): "mymaster",
    }
)
def test_result_backend_transport_options_with_multiple_options():
    """Test that multiple result backend transport options are correctly loaded."""
    import importlib

    # reload celery conf to apply the new config
    importlib.reload(default_celery)
    result_backend_opts = default_celery.DEFAULT_CELERY_CONFIG["result_backend_transport_options"]
    assert result_backend_opts["sentinel_kwargs"] == {"password": "redis_password"}
    assert result_backend_opts["master_name"] == "mymaster"


@conf_vars({("celery_result_backend_transport_options", "sentinel_kwargs"): "invalid_json"})
def test_result_backend_sentinel_kwargs_invalid_json():
    """Test that invalid JSON in sentinel_kwargs raises an error."""
    import importlib

    from airflow.providers.common.compat.sdk import AirflowException

    with pytest.raises(
        AirflowException, match="sentinel_kwargs.*should be written in the correct dictionary format"
    ):
        importlib.reload(default_celery)


@conf_vars({("celery_result_backend_transport_options", "sentinel_kwargs"): '"not_a_dict"'})
def test_result_backend_sentinel_kwargs_not_dict():
    """Test that non-dict sentinel_kwargs raises an error."""
    import importlib

    from airflow.providers.common.compat.sdk import AirflowException

    with pytest.raises(
        AirflowException, match="sentinel_kwargs.*should be written in the correct dictionary format"
    ):
        importlib.reload(default_celery)


@conf_vars(
    {
        ("celery", "result_backend"): "sentinel://sentinel1:26379;sentinel://sentinel2:26379",
        ("celery_result_backend_transport_options", "sentinel_kwargs"): '{"password": "redis_pass"}',
        ("celery_result_backend_transport_options", "master_name"): "mymaster",
    }
)
def test_result_backend_sentinel_full_config():
    """Test full Redis Sentinel configuration for result backend."""
    import importlib

    # reload celery conf to apply the new config
    importlib.reload(default_celery)

    assert default_celery.DEFAULT_CELERY_CONFIG["result_backend"] == (
        "sentinel://sentinel1:26379;sentinel://sentinel2:26379"
    )
    result_backend_opts = default_celery.DEFAULT_CELERY_CONFIG["result_backend_transport_options"]
    assert result_backend_opts["sentinel_kwargs"] == {"password": "redis_pass"}
    assert result_backend_opts["master_name"] == "mymaster"


@pytest.mark.skipif(not AIRFLOW_V_3_2_PLUS, reason="Multi-team support requires Airflow 3.2+")
class TestMultiTeamCeleryExecutor:
    """Test multi-team functionality in CeleryExecutor."""

    def setup_method(self) -> None:
        db.clear_db_runs()
        db.clear_db_jobs()

    def teardown_method(self) -> None:
        db.clear_db_runs()
        db.clear_db_jobs()

    @conf_vars(
        {
            ("celery", "broker_url"): "redis://global:6379/0",
            ("operators", "default_queue"): "global_queue",
        }
    )
    def test_multi_team_isolation_and_task_routing(self, monkeypatch):
        """
        Test multi-team executor isolation and correct task routing.

        Verifies:
        - Each executor has isolated Celery app and config
        - Tasks are routed through team-specific apps (_process_tasks/_process_workloads)
        - Backward compatibility with global executor
        """
        # Set up team-specific config via environment variables
        monkeypatch.setenv("AIRFLOW__TEAM_A___CELERY__BROKER_URL", "redis://team-a:6379/0")
        monkeypatch.setenv("AIRFLOW__TEAM_A___OPERATORS__DEFAULT_QUEUE", "team_a_queue")
        monkeypatch.setenv("AIRFLOW__TEAM_B___CELERY__BROKER_URL", "redis://team-b:6379/0")
        monkeypatch.setenv("AIRFLOW__TEAM_B___OPERATORS__DEFAULT_QUEUE", "team_b_queue")

        # Reload config to pick up environment variables
        from airflow import configuration

        configuration.conf.read_dict({}, source="test")

        # Create executors with different team configs
        team_a_executor = CeleryExecutor(parallelism=2, team_name="team_a")
        team_b_executor = CeleryExecutor(parallelism=3, team_name="team_b")
        global_executor = CeleryExecutor(parallelism=4)

        # Each executor has its own Celery app (critical for isolation)
        assert team_a_executor.celery_app is not team_b_executor.celery_app
        assert team_a_executor.celery_app is not global_executor.celery_app

        # Team-specific broker URLs are used
        assert "team-a" in team_a_executor.celery_app.conf.broker_url
        assert "team-b" in team_b_executor.celery_app.conf.broker_url
        assert "global" in global_executor.celery_app.conf.broker_url

        # Team-specific queues are used
        assert team_a_executor.celery_app.conf.task_default_queue == "team_a_queue"
        assert team_b_executor.celery_app.conf.task_default_queue == "team_b_queue"
        assert global_executor.celery_app.conf.task_default_queue == "global_queue"

        # Each executor has its own BulkStateFetcher with correct app
        assert team_a_executor.bulk_state_fetcher.celery_app is team_a_executor.celery_app
        assert team_b_executor.bulk_state_fetcher.celery_app is team_b_executor.celery_app

        # Executors have isolated internal state
        assert team_a_executor.tasks is not team_b_executor.tasks
        assert team_a_executor.running is not team_b_executor.running
        assert team_a_executor.queued_tasks is not team_b_executor.queued_tasks

    @conf_vars({("celery", "broker_url"): "redis://global:6379/0"})
    @mock.patch("airflow.providers.celery.executors.celery_executor.CeleryExecutor._send_tasks")
    def test_task_routing_through_team_specific_app(self, mock_send_tasks, monkeypatch):
        """
        Test that _process_tasks and _process_workloads pass the correct team_name for task routing.

        With the ProcessPoolExecutor approach, we pass team_name instead of task objects to avoid
        pickling issues. The subprocess reconstructs the team-specific Celery app from the team_name.
        """
        # Set up team A config
        monkeypatch.setenv("AIRFLOW__TEAM_A___CELERY__BROKER_URL", "redis://team-a:6379/0")

        team_a_executor = CeleryExecutor(parallelism=2, team_name="team_a")

        if AIRFLOW_V_3_0_PLUS:
            from airflow.executors.workloads import ExecuteTask
            from airflow.models.taskinstancekey import TaskInstanceKey

            # Create mock workload
            mock_ti = mock.Mock()
            mock_ti.key = TaskInstanceKey("dag", "task", "run", 1)
            mock_ti.queue = "test_queue"
            mock_workload = mock.Mock(spec=ExecuteTask)
            mock_workload.ti = mock_ti

            # Process workload through team A executor
            team_a_executor._process_workloads([mock_workload])

            # Verify _send_tasks received the correct team_name
            assert mock_send_tasks.called
            task_tuples = mock_send_tasks.call_args[0][0]
            team_name_from_call = task_tuples[0][3]  # 4th element is now team_name

            # Critical: team_name is passed so subprocess can reconstruct the correct app
            assert team_name_from_call == "team_a"
        else:
            from airflow.models.taskinstancekey import TaskInstanceKey

            # Test V2 path with execute_command
            mock_key = TaskInstanceKey("dag", "task", "run", 1)
            mock_command = ["airflow", "tasks", "run", "dag", "task"]
            mock_queue = "test_queue"

            # Process task through team A executor
            team_a_executor._process_tasks([(mock_key, mock_command, mock_queue, None)])

            # Verify _send_tasks received team A's execute_command task
            assert mock_send_tasks.called
            task_tuples = mock_send_tasks.call_args[0][0]
            task_from_call = task_tuples[0][3]  # 4th element is the task (V2 still uses task object)

            # Critical: task belongs to team A's app, not module-level app
            assert task_from_call.app is team_a_executor.celery_app
            assert task_from_call.name == "execute_command"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import importlib

import pytest

from airflow.cli import cli_parser
from airflow.providers.celery.cli.definition import CELERY_CLI_COMMANDS, CELERY_COMMANDS

from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.version_compat import AIRFLOW_V_3_2_PLUS


class TestCeleryCliDefinition:
    @pytest.fixture(autouse=True)
    def setup_parser(self):
        if AIRFLOW_V_3_2_PLUS:
            importlib.reload(cli_parser)
            cli_parser.get_parser.cache_clear()
            self.arg_parser = cli_parser.get_parser()
        else:
            with conf_vars(
                {
                    (
                        "core",
                        "executor",
                    ): "CeleryExecutor",
                }
            ):
                importlib.reload(cli_parser)
                cli_parser.get_parser.cache_clear()
                self.arg_parser = cli_parser.get_parser()

    def test_celery_cli_commands_count(self):
        """Test that CELERY_CLI_COMMANDS contains exactly 1 GroupCommand."""
        assert len(CELERY_CLI_COMMANDS) == 1

    def test_celery_commands_count(self):
        """Test that CELERY_COMMANDS contains all 9 subcommands."""
        assert len(CELERY_COMMANDS) == 9

    @pytest.mark.parametrize(
        "command",
        [
            "worker",
            "flower",
            "stop",
            "list-workers",
            "shutdown-worker",
            "shutdown-all-workers",
            "add-queue",
            "remove-queue",
            "remove-all-queues",
        ],
    )
    def test_celery_subcommands_defined(self, command):
        """Test that all celery subcommands are properly defined."""
        params = ["celery", command, "--help"]
        with pytest.raises(SystemExit) as exc_info:
            self.arg_parser.parse_args(params)
        # --help exits with code 0
        assert exc_info.value.code == 0

    def test_worker_command_args(self):
        """Test worker command with various arguments."""
        params = [
            "celery",
            "worker",
            "--queues",
            "queue1,queue2",
            "--concurrency",
            "4",
            "--celery-hostname",
            "worker1",
        ]
        args = self.arg_parser.parse_args(params)
        assert args.queues == "queue1,queue2"
        assert args.concurrency == 4
        assert args.celery_hostname == "worker1"

    def test_flower_command_args(self):
        """Test flower command with various arguments."""
        params = [
            "celery",
            "flower",
            "--hostname",
            "localhost",
            "--port",
            "5555",
            "--url-prefix",
            "/flower",
        ]
        args = self.arg_parser.parse_args(params)
        assert args.hostname == "localhost"
        assert args.port == 5555
        assert args.url_prefix == "/flower"

    def test_list_workers_command_args(self):
        """Test list-workers command with output format."""
        params = ["celery", "list-workers", "--output", "json"]
        args = self.arg_parser.parse_args(params)
        assert args.output == "json"

    def test_shutdown_worker_command_args(self):
        """Test shutdown-worker command with celery hostname."""
        params = ["celery", "shutdown-worker", "--celery-hostname", "celery@worker1"]
        args = self.arg_parser.parse_args(params)
        assert args.celery_hostname == "celery@worker1"

    def test_shutdown_all_workers_command_args(self):
        """Test shutdown-all-workers command with yes flag."""
        params = ["celery", "shutdown-all-workers", "--yes"]
        args = self.arg_parser.parse_args(params)
        assert args.yes is True

    def test_add_queue_command_args(self):
        """Test add-queue command with required arguments."""
        params = [
            "celery",
            "add-queue",
            "--queues",
            "new_queue",
            "--celery-hostname",
            "celery@worker1",
        ]
        args = self.arg_parser.parse_args(params)
        assert args.queues == "new_queue"
        assert args.celery_hostname == "celery@worker1"

    def test_remove_queue_command_args(self):
        """Test remove-queue command with required arguments."""
        params = [
            "celery",
            "remove-queue",
            "--queues",
            "old_queue",
            "--celery-hostname",
            "celery@worker1",
        ]
        args = self.arg_parser.parse_args(params)
        assert args.queues == "old_queue"
        assert args.celery_hostname == "celery@worker1"

    def test_remove_all_queues_command_args(self):
        """Test remove-all-queues command with celery hostname."""
        params = ["celery", "remove-all-queues", "--celery-hostname", "celery@worker1"]
        args = self.arg_parser.parse_args(params)
        assert args.celery_hostname == "celery@worker1"

    def test_stop_command_args(self):
        """Test stop command with pid argument."""
        params = ["celery", "stop", "--pid", "/path/to/pid"]
        args = self.arg_parser.parse_args(params)
        assert args.pid == "/path/to/pid"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from celery.app import control

from airflow.providers.celery.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import BaseSensorOperator
else:
    from airflow.sensors.base import BaseSensorOperator  # type: ignore[no-redef]

if TYPE_CHECKING:
    from airflow.sdk import Context


class CeleryQueueSensor(BaseSensorOperator):
    """
    Waits for a Celery queue to be empty.

    By default, in order to be considered empty, the queue must not have
    any tasks in the ``reserved``, ``scheduled`` or ``active`` states.

    :param celery_queue: The name of the Celery queue to wait for.
    :param target_task_id: Task id for checking
    """

    def __init__(self, *, celery_queue: str, target_task_id: str | None = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.celery_queue = celery_queue
        self.target_task_id = target_task_id

    def _check_task_id(self, context: Context) -> bool:
        """
        Get the Celery result from the Airflow task ID and return True if the result has finished execution.

        :param context: Airflow's execution context
        :return: True if task has been executed, otherwise False
        """
        ti = context["ti"]
        celery_result = ti.xcom_pull(task_ids=self.target_task_id)
        return celery_result.ready()

    def poke(self, context: Context) -> bool:
        if self.target_task_id:
            return self._check_task_id(context)

        inspect_result = control.Inspect()
        reserved = inspect_result.reserved()
        scheduled = inspect_result.scheduled()
        active = inspect_result.active()

        try:
            reserved = len(reserved[self.celery_queue])
            scheduled = len(scheduled[self.celery_queue])
            active = len(active[self.celery_queue])

            self.log.info("Checking if celery queue %s is empty.", self.celery_queue)

            return reserved == 0 and scheduled == 0 and active == 0
        except KeyError:
            message = f"Could not locate Celery queue {self.celery_queue}"
            raise KeyError(message)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib
import importlib
import json
import logging
import os
import sys
from io import StringIO
from unittest import mock
from unittest.mock import MagicMock, patch

import pytest

from airflow.cli import cli_parser
from airflow.configuration import conf
from airflow.executors import executor_loader
from airflow.providers.celery.cli import celery_command
from airflow.providers.celery.cli.celery_command import _run_stale_bundle_cleanup

from tests_common.test_utils.config import conf_vars
from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS, AIRFLOW_V_3_2_PLUS

PY313 = sys.version_info >= (3, 13)


@pytest.fixture(autouse=False)
def conf_stale_bundle_cleanup_disabled():
    with conf_vars({("dag_processor", "stale_bundle_cleanup_interval"): "0"}):
        yield


@pytest.mark.backend("mysql", "postgres")
@pytest.mark.usefixtures("conf_stale_bundle_cleanup_disabled")
class TestCeleryStopCommand:
    @classmethod
    def setup_class(cls):
        with conf_vars({("core", "executor"): "CeleryExecutor"}):
            importlib.reload(executor_loader)
            importlib.reload(cli_parser)
            cls.parser = cli_parser.get_parser()

    @mock.patch("airflow.providers.celery.cli.celery_command.setup_locations")
    @mock.patch("airflow.providers.celery.cli.celery_command.psutil.Process")
    def test_if_right_pid_is_read(self, mock_process, mock_setup_locations, tmp_path):
        args = self.parser.parse_args(["celery", "stop"])
        pid = "123"
        path = tmp_path / "testfile"
        # Create pid file
        path.write_text(pid)
        # Setup mock
        mock_setup_locations.return_value = (os.fspath(path), None, None, None)

        # Calling stop_worker should delete the temporary pid file
        celery_command.stop_worker(args)
        # Check if works as expected
        assert not path.exists()
        mock_process.assert_called_once_with(int(pid))
        mock_process.return_value.terminate.assert_called_once_with()

    @mock.patch("airflow.providers.celery.cli.celery_command.read_pid_from_pidfile")
    @mock.patch("airflow.providers.celery.executors.celery_executor.app")
    @mock.patch("airflow.providers.celery.cli.celery_command.setup_locations")
    def test_same_pid_file_is_used_in_start_and_stop(
        self, mock_setup_locations, mock_celery_app, mock_read_pid_from_pidfile
    ):
        pid_file = "test_pid_file"
        mock_setup_locations.return_value = (pid_file, None, None, None)
        mock_read_pid_from_pidfile.return_value = None

        # Call worker
        worker_args = self.parser.parse_args(["celery", "worker", "--skip-serve-logs"])
        celery_command.worker(worker_args)
        assert mock_celery_app.worker_main.call_args
        args, _ = mock_celery_app.worker_main.call_args
        args_str = " ".join(map(str, args[0]))
        assert f"--pidfile {pid_file}" not in args_str

        # Call stop
        stop_args = self.parser.parse_args(["celery", "stop"])
        celery_command.stop_worker(stop_args)
        mock_read_pid_from_pidfile.assert_called_once_with(pid_file)

    @mock.patch("airflow.providers.celery.cli.celery_command.remove_existing_pidfile")
    @mock.patch("airflow.providers.celery.cli.celery_command.read_pid_from_pidfile")
    @mock.patch("airflow.providers.celery.executors.celery_executor.app")
    @mock.patch("airflow.providers.celery.cli.celery_command.psutil.Process")
    @mock.patch("airflow.providers.celery.cli.celery_command.setup_locations")
    def test_custom_pid_file_is_used_in_start_and_stop(
        self,
        mock_setup_locations,
        mock_process,
        mock_celery_app,
        mock_read_pid_from_pidfile,
        mock_remove_existing_pidfile,
    ):
        pid_file = "custom_test_pid_file"
        mock_setup_locations.return_value = (pid_file, None, None, None)
        # Call worker
        worker_args = self.parser.parse_args(["celery", "worker", "--skip-serve-logs", "--pid", pid_file])
        celery_command.worker(worker_args)
        assert mock_celery_app.worker_main.call_args
        args, _ = mock_celery_app.worker_main.call_args
        args_str = " ".join(map(str, args[0]))
        assert f"--pidfile {pid_file}" not in args_str

        stop_args = self.parser.parse_args(["celery", "stop", "--pid", pid_file])
        celery_command.stop_worker(stop_args)

        mock_read_pid_from_pidfile.assert_called_once_with(pid_file)
        mock_process.return_value.terminate.assert_called()
        mock_remove_existing_pidfile.assert_called_once_with(pid_file)


@pytest.mark.backend("mysql", "postgres")
@pytest.mark.usefixtures("conf_stale_bundle_cleanup_disabled")
class TestWorkerStart:
    @classmethod
    def setup_class(cls):
        with conf_vars({("core", "executor"): "CeleryExecutor"}):
            importlib.reload(executor_loader)
            importlib.reload(cli_parser)
            cls.parser = cli_parser.get_parser()

    @mock.patch("airflow.providers.celery.cli.celery_command.setup_locations")
    @mock.patch("airflow.providers.celery.cli.celery_command.Process")
    @mock.patch("airflow.providers.celery.executors.celery_executor.app")
    def test_worker_started_with_required_arguments(self, mock_celery_app, mock_popen, mock_locations):
        pid_file = "pid_file"
        mock_locations.return_value = (pid_file, None, None, None)
        concurrency = "1"
        celery_hostname = "celery_hostname"
        queues = "queue"
        autoscale = "2,5"
        args = self.parser.parse_args(
            [
                "celery",
                "worker",
                "--autoscale",
                autoscale,
                "--concurrency",
                concurrency,
                "--celery-hostname",
                celery_hostname,
                "--queues",
                queues,
                "--without-mingle",
                "--without-gossip",
            ]
        )

        celery_command.worker(args)

        mock_celery_app.worker_main.assert_called_once_with(
            [
                "worker",
                "-O",
                "fair",
                "--queues",
                queues,
                "--concurrency",
                int(concurrency),
                "--loglevel",
                conf.get("logging", "CELERY_LOGGING_LEVEL"),
                "--hostname",
                celery_hostname,
                "--autoscale",
                autoscale,
                "--without-mingle",
                "--without-gossip",
                "--pool",
                "prefork",
            ]
        )


@pytest.mark.backend("mysql", "postgres")
@pytest.mark.usefixtures("conf_stale_bundle_cleanup_disabled")
class TestWorkerMultiTeam:
    @classmethod
    def setup_class(cls):
        with conf_vars({("core", "executor"): "CeleryExecutor"}):
            importlib.reload(executor_loader)
            importlib.reload(cli_parser)
            cls.parser = cli_parser.get_parser()

    @pytest.mark.skipif(not AIRFLOW_V_3_2_PLUS, reason="Multi-team requires Airflow 3.2+")
    @mock.patch("airflow.providers.celery.cli.celery_command.conf")
    @mock.patch("airflow.providers.celery.cli.celery_command.setup_locations")
    @mock.patch("airflow.providers.celery.cli.celery_command.Process")
    @mock.patch("airflow.providers.celery.executors.celery_executor_utils.create_celery_app")
    @mock.patch("airflow.executors.base_executor.ExecutorConf")
    def test_worker_with_team_creates_team_specific_app(
        self, mock_executor_conf_class, mock_create_celery_app, mock_popen, mock_locations, mock_global_conf
    ):
        """Test that worker with --team parameter creates team-specific Celery app and uses ExecutorConf."""
        pid_file = "pid_file"
        mock_locations.return_value = (pid_file, None, None, None)

        # Mock global conf for stale bundle cleanup check
        mock_global_conf.getint.return_value = 0

        # Setup mock team config
        mock_team_config = MagicMock()
        mock_team_config.get.side_effect = lambda section, key, **kwargs: {
            ("logging", "CELERY_LOGGING_LEVEL"): "INFO",
            ("celery", "worker_autoscale"): None,
            ("celery", "pool"): "prefork",
            ("celery", "worker_umask"): "0o077",
        }.get((section, key), kwargs.get("fallback"))
        mock_team_config.getint.return_value = 0  # For stale_bundle_cleanup_interval
        mock_team_config.has_option.return_value = True
        mock_executor_conf_class.return_value = mock_team_config

        # Setup mock team-specific Celery app
        mock_team_app = MagicMock()
        mock_team_app.control.inspect.return_value.active_queues.return_value = None
        mock_create_celery_app.return_value = mock_team_app

        # Parse args with --team
        args = self.parser.parse_args(
            [
                "celery",
                "worker",
                "--team",
                "team_a",
                "--queues",
                "team_a_queue",
                "--concurrency",
                "2",
                "--skip-serve-logs",
            ]
        )

        # Call worker
        celery_command.worker(args)

        # Verify ExecutorConf was created with correct team name
        mock_executor_conf_class.assert_called_once_with(team_name="team_a")

        # Verify create_celery_app was called with team config
        mock_create_celery_app.assert_called_once_with(mock_team_config)

        # Verify team-specific app was used for worker_main
        mock_team_app.worker_main.assert_called_once()
        args_list = mock_team_app.worker_main.call_args[0][0]
        assert "--queues" in args_list
        assert "team_a_queue" in args_list

        # Verify team config was used for configuration reads
        # We expect at least 4 config reads: CELERY_LOGGING_LEVEL, LOGGING_LEVEL, pool, worker_umask
        assert mock_team_config.get.call_count >= 4
        assert mock_team_config.has_option.called

        # Verify global conf was NOT used for these worker-specific config reads
        # The global conf should not be accessed for celery or logging config in multi-team mode
        global_conf_calls = [
            call for call in mock_global_conf.get.call_args_list if call[0][0] in ("celery", "logging")
        ]
        assert len(global_conf_calls) == 0, (
            f"Global conf should not be used in multi-team mode, but was called: {global_conf_calls}"
        )

    # Test which covers the scenario of a new provider that supports teams running on an older version of
    # Airflow. It should be run only on Airflow versions below 3.2 and above or equal to 2.11. We should
    # expect the worker to throw an Airflow configuration exception.
    @pytest.mark.skipif(AIRFLOW_V_3_2_PLUS or not AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 2.11-3.1")
    def test_worker_with_team_raises_on_older_airflow_versions(self):
        args = self.parser.parse_args(
            [
                "celery",
                "worker",
                "--team",
                "team_a",
            ]
        )

        with pytest.raises(SystemExit) as exc_info:
            celery_command.worker(args)

        assert "Multi-team Celery workers require Airflow version 3.2 or higher" in str(exc_info.value)


@pytest.mark.backend("mysql", "postgres")
@pytest.mark.usefixtures("conf_stale_bundle_cleanup_disabled")
class TestWorkerFailure:
    @classmethod
    def setup_class(cls):
        with conf_vars({("core", "executor"): "CeleryExecutor"}):
            importlib.reload(cli_parser)
            cls.parser = cli_parser.get_parser()

    @mock.patch("airflow.providers.celery.cli.celery_command.Process")
    @mock.patch("airflow.providers.celery.executors.celery_executor.app")
    def test_worker_failure_gracefull_shutdown(self, mock_celery_app, mock_popen):
        args = self.parser.parse_args(["celery", "worker"])
        mock_celery_app.run.side_effect = Exception("Mock exception to trigger runtime error")
        try:
            celery_command.worker(args)
        finally:
            mock_popen().terminate.assert_called()


@pytest.mark.backend("mysql", "postgres")
@pytest.mark.usefixtures("conf_stale_bundle_cleanup_disabled")
class TestWorkerDuplicateHostnameCheck:
    @classmethod
    def setup_class(cls):
        with conf_vars({("core", "executor"): "CeleryExecutor"}):
            importlib.reload(executor_loader)
            importlib.reload(cli_parser)
            cls.parser = cli_parser.get_parser()

    @pytest.mark.db_test
    @mock.patch("airflow.providers.celery.executors.celery_executor.app.control.inspect")
    def test_worker_fails_when_hostname_already_exists(self, mock_inspect):
        """Test that worker command fails when trying to start a worker with a duplicate hostname."""
        args = self.parser.parse_args(["celery", "worker", "--celery-hostname", "existing_host"])

        # Mock the inspect to return an active worker with the same hostname
        mock_instance = MagicMock()
        mock_instance.active_queues.return_value = {
            "celery@existing_host": [{"name": "queue1"}],
        }
        mock_inspect.return_value = mock_instance

        # Test that SystemExit is raised with appropriate error message
        with pytest.raises(SystemExit) as exc_info:
            celery_command.worker(args)

        assert "existing_host" in str(exc_info.value)
        assert "already running" in str(exc_info.value)

    @pytest.mark.db_test
    @mock.patch("airflow.providers.celery.executors.celery_executor.app.control.inspect")
    @mock.patch("airflow.providers.celery.cli.celery_command.Process")
    @mock.patch("airflow.providers.celery.executors.celery_executor.app")
    def test_worker_starts_when_hostname_is_unique(self, mock_celery_app, mock_popen, mock_inspect):
        """Test that worker command succeeds when the hostname is unique."""
        args = self.parser.parse_args(["celery", "worker", "--celery-hostname", "new_host"])

        # Mock the inspect to return active workers without the new hostname
        mock_instance = MagicMock()
        mock_instance.active_queues.return_value = {
            "celery@existing_host": [{"name": "queue1"}],
        }
        mock_inspect.return_value = mock_instance

        # Worker should start successfully
        celery_command.worker(args)

        # Verify that worker_main was called
        assert mock_celery_app.worker_main.called


@pytest.mark.backend("mysql", "postgres")
@pytest.mark.usefixtures("conf_stale_bundle_cleanup_disabled")
class TestFlowerCommand:
    @classmethod
    def setup_class(cls):
        with conf_vars({("core", "executor"): "CeleryExecutor"}):
            importlib.reload(cli_parser)
            cls.parser = cli_parser.get_parser()

    @mock.patch("airflow.providers.celery.executors.celery_executor.app")
    def test_run_command(self, mock_celery_app):
        args = self.parser.parse_args(
            [
                "celery",
                "flower",
                "--basic-auth",
                "admin:admin",
                "--broker-api",
                "http://username:password@rabbitmq-server-name:15672/api/",
                "--flower-conf",
                "flower_config",
                "--hostname",
                "my-hostname",
                "--port",
                "3333",
                "--url-prefix",
                "flower-monitoring",
            ]
        )

        celery_command.flower(args)
        mock_celery_app.start.assert_called_once_with(
            [
                "flower",
                conf.get("celery", "BROKER_URL"),
                "--address=my-hostname",
                "--port=3333",
                "--broker-api=http://username:password@rabbitmq-server-name:15672/api/",
                "--url-prefix=flower-monitoring",
                "--basic-auth=admin:admin",
                "--conf=flower_config",
            ]
        )

    def _test_run_command_daemon(self, mock_celery_app, mock_daemon, mock_setup_locations, mock_pid_file):
        mock_setup_locations.return_value = (
            mock.MagicMock(name="pidfile"),
            mock.MagicMock(name="stdout"),
            mock.MagicMock(name="stderr"),
            mock.MagicMock(name="INVALID"),
        )
        args = self.parser.parse_args(
            [
                "celery",
                "flower",
                "--basic-auth",
                "admin:admin",
                "--broker-api",
                "http://username:password@rabbitmq-server-name:15672/api/",
                "--flower-conf",
                "flower_config",
                "--hostname",
                "my-hostname",
                "--log-file",
                "/tmp/flower.log",
                "--pid",
                "/tmp/flower.pid",
                "--port",
                "3333",
                "--stderr",
                "/tmp/flower-stderr.log",
                "--stdout",
                "/tmp/flower-stdout.log",
                "--url-prefix",
                "flower-monitoring",
                "--daemon",
            ]
        )
        mock_open = mock.mock_open()
        with mock.patch("airflow.cli.commands.daemon_utils.open", mock_open):
            celery_command.flower(args)

        mock_celery_app.start.assert_called_once_with(
            [
                "flower",
                conf.get("celery", "BROKER_URL"),
                "--address=my-hostname",
                "--port=3333",
                "--broker-api=http://username:password@rabbitmq-server-name:15672/api/",
                "--url-prefix=flower-monitoring",
                "--basic-auth=admin:admin",
                "--conf=flower_config",
            ]
        )
        assert mock_daemon.mock_calls[:3] == [
            mock.call.DaemonContext(
                pidfile=mock_pid_file.return_value,
                files_preserve=None,
                stdout=mock_open.return_value,
                stderr=mock_open.return_value,
                umask=0o077,
            ),
            mock.call.DaemonContext().__enter__(),
            mock.call.DaemonContext().__exit__(None, None, None),
        ]

        assert mock_setup_locations.mock_calls == [
            mock.call(
                process="flower",
                pid="/tmp/flower.pid",
                stdout="/tmp/flower-stdout.log",
                stderr="/tmp/flower-stderr.log",
                log="/tmp/flower.log",
            )
        ]
        mock_pid_file.assert_has_calls([mock.call(mock_setup_locations.return_value[0], -1)])

        if PY313:
            assert mock_open.mock_calls == [
                mock.call(mock_setup_locations.return_value[1], "a"),
                mock.call().__enter__(),
                mock.call(mock_setup_locations.return_value[2], "a"),
                mock.call().__enter__(),
                mock.call().truncate(0),
                mock.call().truncate(0),
                mock.call().__exit__(None, None, None),
                mock.call().close(),
                mock.call().__exit__(None, None, None),
                mock.call().close(),
            ]
        else:
            assert mock_open.mock_calls == [
                mock.call(mock_setup_locations.return_value[1], "a"),
                mock.call().__enter__(),
                mock.call(mock_setup_locations.return_value[2], "a"),
                mock.call().__enter__(),
                mock.call().truncate(0),
                mock.call().truncate(0),
                mock.call().__exit__(None, None, None),
                mock.call().__exit__(None, None, None),
            ]

    @pytest.mark.skipif(AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 3.0-")
    @mock.patch("airflow.cli.commands.daemon_utils.TimeoutPIDLockFile")
    @mock.patch("airflow.cli.commands.daemon_utils.setup_locations")
    @mock.patch("airflow.cli.commands.daemon_utils.daemon")
    @mock.patch("airflow.providers.celery.executors.celery_executor.app")
    def test_run_command_daemon_v_3_below(
        self, mock_celery_app, mock_daemon, mock_setup_locations, mock_pid_file
    ):
        self._test_run_command_daemon(mock_celery_app, mock_daemon, mock_setup_locations, mock_pid_file)

    @pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Test requires Airflow 3.0+")
    @mock.patch("airflow.cli.commands.daemon_utils.TimeoutPIDLockFile")
    @mock.patch("airflow.cli.commands.daemon_utils.setup_locations")
    @mock.patch("airflow.cli.commands.daemon_utils.daemon")
    @mock.patch("airflow.providers.celery.executors.celery_executor.app")
    def test_run_command_daemon_v3_above(
        self, mock_celery_app, mock_daemon, mock_setup_locations, mock_pid_file
    ):
        self._test_run_command_daemon(mock_celery_app, mock_daemon, mock_setup_locations, mock_pid_file)


class TestRemoteCeleryControlCommands:
    @classmethod
    def setup_class(cls):
        with conf_vars({("core", "executor"): "CeleryExecutor"}):
            importlib.reload(executor_loader)
            importlib.reload(cli_parser)
            cls.parser = cli_parser.get_parser()

    @pytest.mark.db_test
    @mock.patch("airflow.providers.celery.executors.celery_executor.app.control.inspect")
    def test_list_celery_workers(self, mock_inspect):
        args = self.parser.parse_args(["celery", "list-workers", "--output", "json"])
        mock_instance = MagicMock()
        mock_instance.active_queues.return_value = {
            "celery@host_1": [{"name": "queue1"}, {"name": "queue2"}],
            "celery@host_2": [{"name": "queue3"}],
        }
        mock_inspect.return_value = mock_instance
        with contextlib.redirect_stdout(StringIO()) as temp_stdout:
            celery_command.list_workers(args)
            out = temp_stdout.getvalue()
            celery_workers = json.loads(out)
        for key in ["worker_name", "queues"]:
            assert key in celery_workers[0]
        assert any("celery@host_1" in h["worker_name"] for h in celery_workers)

    @pytest.mark.db_test
    @mock.patch("airflow.providers.celery.executors.celery_executor.app.control.shutdown")
    def test_shutdown_worker(self, mock_shutdown):
        args = self.parser.parse_args(["celery", "shutdown-worker", "-H", "celery@host_1"])
        with patch(
            "airflow.providers.celery.cli.celery_command._check_if_active_celery_worker", return_value=None
        ):
            celery_command.shutdown_worker(args)
            mock_shutdown.assert_called_once_with(destination=["celery@host_1"])

    @pytest.mark.db_test
    @mock.patch("airflow.providers.celery.executors.celery_executor.app.control.broadcast")
    def test_shutdown_all_workers(self, mock_broadcast):
        args = self.parser.parse_args(["celery", "shutdown-all-workers", "-y"])
        with patch(
            "airflow.providers.celery.cli.celery_command._check_if_active_celery_worker", return_value=None
        ):
            celery_command.shutdown_all_workers(args)
            mock_broadcast.assert_called_once_with("shutdown")

    @pytest.mark.db_test
    @mock.patch("airflow.providers.celery.executors.celery_executor.app.control.add_consumer")
    def test_add_queue(self, mock_add_consumer):
        args = self.parser.parse_args(["celery", "add-queue", "-q", "test1", "-H", "celery@host_1"])
        with patch(
            "airflow.providers.celery.cli.celery_command._check_if_active_celery_worker", return_value=None
        ):
            celery_command.add_queue(args)
            mock_add_consumer.assert_called_once_with("test1", destination=["celery@host_1"])

    @pytest.mark.db_test
    @mock.patch("airflow.providers.celery.executors.celery_executor.app.control.cancel_consumer")
    def test_remove_queue(self, mock_cancel_consumer):
        args = self.parser.parse_args(["celery", "remove-queue", "-q", "test1", "-H", "celery@host_1"])
        with patch(
            "airflow.providers.celery.cli.celery_command._check_if_active_celery_worker", return_value=None
        ):
            celery_command.remove_queue(args)
            mock_cancel_consumer.assert_called_once_with("test1", destination=["celery@host_1"])

    @pytest.mark.db_test
    @mock.patch("airflow.providers.celery.executors.celery_executor.app.control.cancel_consumer")
    @mock.patch("airflow.providers.celery.executors.celery_executor.app.control.inspect")
    def test_remove_all_queues(self, mock_inspect, mock_cancel_consumer):
        args = self.parser.parse_args(["celery", "remove-all-queues", "-H", "celery@host_1"])
        mock_instance = MagicMock()
        mock_instance.active_queues.return_value = {
            "celery@host_1": [{"name": "queue1"}, {"name": "queue2"}],
            "celery@host_2": [{"name": "queue3"}],
        }
        mock_inspect.return_value = mock_instance
        with patch(
            "airflow.providers.celery.cli.celery_command._check_if_active_celery_worker", return_value=None
        ):
            celery_command.remove_all_queues(args)
            # Verify cancel_consumer was called for each queue
            expected_calls = [
                mock.call("queue1", destination=["celery@host_1"]),
                mock.call("queue2", destination=["celery@host_1"]),
            ]
            mock_cancel_consumer.assert_has_calls(expected_calls, any_order=True)
            assert mock_cancel_consumer.call_count == 2


@patch("airflow.providers.celery.cli.celery_command.Process")
@pytest.mark.skipif(not AIRFLOW_V_3_0_PLUS, reason="Doesn't apply to pre-3.0")
def test_stale_bundle_cleanup(mock_process):
    mock_process.__bool__.return_value = True
    with _run_stale_bundle_cleanup():
        ...
    calls = mock_process.call_args_list
    assert len(calls) == 1
    actual = [x.kwargs["target"] for x in calls]
    assert actual[0].__name__ == "bundle_cleanup_main"


class TestLoggerSetupHandler:
    """Tests for logger_setup_handler that configures celery logging."""

    def test_logger_setup_handler_uses_task_formatter_when_separation_enabled(self):
        """When celery_stdout_stderr_separation is True, handlers should use TaskFormatter."""
        from celery.app.log import TaskFormatter

        logger = MagicMock(spec=logging.Logger)
        logger.handlers = []

        with conf_vars({("logging", "celery_stdout_stderr_separation"): "True"}):
            celery_command.logger_setup_handler(logger)

        assert len(logger.handlers) == 2
        stdout_handler, stderr_handler = logger.handlers
        assert isinstance(stdout_handler.formatter, TaskFormatter)
        assert isinstance(stderr_handler.formatter, TaskFormatter)

    def test_logger_setup_handler_stdout_stderr_split(self):
        """Verify stdout handler filters errors out, stderr handler only gets errors and above."""
        logger = MagicMock(spec=logging.Logger)
        logger.handlers = []

        with conf_vars({("logging", "celery_stdout_stderr_separation"): "True"}):
            celery_command.logger_setup_handler(logger)

        stdout_handler, stderr_handler = logger.handlers

        # stdout handler should have a filter that rejects ERROR and above
        assert len(stdout_handler.filters) == 1
        error_record = logging.LogRecord("test", logging.ERROR, "", 0, "msg", (), None)
        warning_record = logging.LogRecord("test", logging.WARNING, "", 0, "msg", (), None)
        assert stdout_handler.filters[0].filter(error_record) is False
        assert stdout_handler.filters[0].filter(warning_record) is True

        # stderr handler level should be ERROR
        assert stderr_handler.level == logging.ERROR

    def test_logger_setup_handler_noop_when_separation_disabled(self):
        """When celery_stdout_stderr_separation is False, logger handlers should not be modified."""
        logger = MagicMock(spec=logging.Logger)
        original_handlers = [logging.StreamHandler()]
        logger.handlers = original_handlers.copy()

        with conf_vars({("logging", "celery_stdout_stderr_separation"): "False"}):
            celery_command.logger_setup_handler(logger)

        # Handlers should remain unchanged
        assert logger.handlers == original_handlers



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Celery command."""

from __future__ import annotations

import logging
import sys
import time
from contextlib import contextmanager, suppress
from multiprocessing import Process

import psutil
import sqlalchemy.exc
from celery import maybe_patch_concurrency
from celery.app.defaults import DEFAULT_TASK_LOG_FMT
from celery.app.log import TaskFormatter
from celery.signals import after_setup_logger
from lockfile.pidlockfile import read_pid_from_pidfile, remove_existing_pidfile

from airflow import settings
from airflow.cli.simple_table import AirflowConsole
from airflow.configuration import conf
from airflow.exceptions import AirflowConfigException
from airflow.providers.celery.version_compat import AIRFLOW_V_3_0_PLUS, AIRFLOW_V_3_2_PLUS
from airflow.utils import cli as cli_utils
from airflow.utils.cli import setup_locations

WORKER_PROCESS_NAME = "worker"

log = logging.getLogger(__name__)


def _run_command_with_daemon_option(*args, **kwargs):
    try:
        from airflow.cli.commands.daemon_utils import run_command_with_daemon_option

        run_command_with_daemon_option(*args, **kwargs)
    except ImportError:
        from airflow.providers.common.compat.sdk import AirflowOptionalProviderFeatureException

        raise AirflowOptionalProviderFeatureException(
            "Failed to import run_command_with_daemon_option. This feature is only available in Airflow versions >= 2.8.0"
        )


def _providers_configuration_loaded(func):
    def wrapper(*args, **kwargs):
        try:
            from airflow.utils.providers_configuration_loader import providers_configuration_loaded

            providers_configuration_loaded(func)(*args, **kwargs)
        except ImportError as e:
            from airflow.providers.common.compat.sdk import AirflowOptionalProviderFeatureException

            raise AirflowOptionalProviderFeatureException(
                "Failed to import providers_configuration_loaded. This feature is only available in Airflow versions >= 2.8.0"
            ) from e

    return wrapper


@cli_utils.action_cli(check_db=False)
@_providers_configuration_loaded
def flower(args):
    """Start Flower, Celery monitoring tool."""
    # This needs to be imported locally to not trigger Providers Manager initialization
    from airflow.providers.celery.executors.celery_executor import app as celery_app

    options = [
        "flower",
        conf.get("celery", "BROKER_URL"),
        f"--address={args.hostname}",
        f"--port={args.port}",
    ]

    if args.broker_api:
        options.append(f"--broker-api={args.broker_api}")

    if args.url_prefix:
        options.append(f"--url-prefix={args.url_prefix}")

    if args.basic_auth:
        options.append(f"--basic-auth={args.basic_auth}")

    if args.flower_conf:
        options.append(f"--conf={args.flower_conf}")

    _run_command_with_daemon_option(
        args=args, process_name="flower", callback=lambda: celery_app.start(options)
    )


@contextmanager
def _serve_logs(skip_serve_logs: bool = False):
    """Start serve_logs sub-process."""
    from airflow.utils.serve_logs import serve_logs

    sub_proc = None
    if skip_serve_logs is False:
        sub_proc = Process(target=serve_logs)
        sub_proc.start()
    try:
        yield
    finally:
        if sub_proc:
            sub_proc.terminate()


@contextmanager
def _run_stale_bundle_cleanup():
    """Start stale bundle cleanup sub-process."""
    check_interval = None
    with suppress(AirflowConfigException):  # remove when min airflow version >= 3.0
        check_interval = conf.getint(
            section="dag_processor",
            key="stale_bundle_cleanup_interval",
        )
    if not check_interval or check_interval <= 0 or not AIRFLOW_V_3_0_PLUS:
        # do not start bundle cleanup process
        with suppress(BaseException):
            yield
        return
    from airflow.dag_processing.bundles.base import BundleUsageTrackingManager

    log.info("starting stale bundle cleanup process")
    sub_proc = None

    def bundle_cleanup_main():
        mgr = BundleUsageTrackingManager()
        while True:
            time.sleep(check_interval)
            mgr.remove_stale_bundle_versions()

    try:
        sub_proc = Process(target=bundle_cleanup_main)
        sub_proc.start()
        yield
    finally:
        if sub_proc:
            sub_proc.terminate()


@after_setup_logger.connect()
@_providers_configuration_loaded
def logger_setup_handler(logger, **kwargs):
    """
    Reconfigure the logger.

    * remove any previously configured handlers
    * logs of severity error, and above goes to stderr,
    * logs of severity lower than error goes to stdout.
    """
    if conf.getboolean("logging", "celery_stdout_stderr_separation", fallback=False):
        celery_formatter = TaskFormatter(DEFAULT_TASK_LOG_FMT)

        class NoErrorOrAboveFilter(logging.Filter):
            """Allow only logs with level *lower* than ERROR to be reported."""

            def filter(self, record):
                return record.levelno < logging.ERROR

        below_error_handler = logging.StreamHandler(sys.stdout)
        below_error_handler.addFilter(NoErrorOrAboveFilter())
        below_error_handler.setFormatter(celery_formatter)

        from_error_handler = logging.StreamHandler(sys.stderr)
        from_error_handler.setLevel(logging.ERROR)
        from_error_handler.setFormatter(celery_formatter)

        logger.handlers[:] = [below_error_handler, from_error_handler]


@cli_utils.action_cli(check_db=not AIRFLOW_V_3_0_PLUS)
@_providers_configuration_loaded
def worker(args):
    """Start Airflow Celery worker."""
    team_config = None
    if hasattr(args, "team") and args.team:
        # Multi-team is enabled, create team-specific Celery app and use team based config
        # This requires Airflow 3.2+, and core.multi_team config to be true to be enabled.
        if not AIRFLOW_V_3_2_PLUS:
            raise SystemExit(
                "Error: Multi-team Celery workers require Airflow version 3.2 or higher. "
                "Please upgrade your Airflow installation or remove the --team argument."
            )
        if not conf.getboolean("core", "multi_team", fallback=False):
            raise SystemExit(
                "Error: Multi-team Celery workers require core.multi_team configuration to be enabled. "
                "Please enable core.multi_team in your Airflow config or remove the --team argument."
            )
        from airflow.executors.base_executor import ExecutorConf
        from airflow.providers.celery.executors.celery_executor_utils import create_celery_app

        team_config = ExecutorConf(team_name=args.team)
        log.info("Starting Celery worker for team: %s", args.team)
        celery_app = create_celery_app(team_config)
    else:
        # Backward compatible: use module-level app with global config
        from airflow.providers.celery.executors.celery_executor import app as celery_app

    # Use team_config for config reads in multi-team mode, otherwise use global conf
    config = team_config if team_config else conf

    # Check if a worker with the same hostname already exists
    if args.celery_hostname:
        inspect = celery_app.control.inspect()
        active_workers = inspect.active_queues()
        if active_workers:
            active_worker_names = list(active_workers.keys())
            # Check if any worker ends with @hostname
            if any(name.endswith(f"@{args.celery_hostname}") for name in active_worker_names):
                raise SystemExit(
                    f"Error: A worker with hostname '{args.celery_hostname}' is already running. "
                    "Please use a different hostname or stop the existing worker first."
                )

    if AIRFLOW_V_3_0_PLUS:
        from airflow.sdk.log import configure_logging

        configure_logging(output=sys.stdout.buffer)
    else:
        # Disable connection pool so that celery worker does not hold an unnecessary db connection
        settings.reconfigure_orm(disable_connection_pool=True)
        if not settings.validate_session():
            raise SystemExit("Worker exiting, database connection precheck failed.")

    autoscale = args.autoscale
    skip_serve_logs = args.skip_serve_logs

    if autoscale is None and config.has_option("celery", "worker_autoscale"):
        autoscale = config.get("celery", "worker_autoscale")

    if hasattr(celery_app.backend, "ResultSession"):
        # Pre-create the database tables now, otherwise SQLA via Celery has a
        # race condition where one of the subprocesses can die with "Table
        # already exists" error, because SQLA checks for which tables exist,
        # then issues a CREATE TABLE, rather than doing CREATE TABLE IF NOT
        # EXISTS
        try:
            session = celery_app.backend.ResultSession()
            session.close()
        except sqlalchemy.exc.IntegrityError:
            # At least on postgres, trying to create a table that already exist
            # gives a unique constraint violation or the
            # "pg_type_typname_nsp_index" table. If this happens we can ignore
            # it, we raced to create the tables and lost.
            pass

    # backwards-compatible: https://github.com/apache/airflow/pull/21506#pullrequestreview-879893763
    celery_log_level = config.get("logging", "CELERY_LOGGING_LEVEL")
    if not celery_log_level:
        celery_log_level = config.get("logging", "LOGGING_LEVEL")

    # Setup Celery worker
    options = [
        "worker",
        "-O",
        "fair",
        "--queues",
        args.queues,
        "--concurrency",
        args.concurrency,
        "--loglevel",
        celery_log_level,
    ]
    if args.celery_hostname:
        options.extend(["--hostname", args.celery_hostname])
    if autoscale:
        options.extend(["--autoscale", autoscale])
    if args.without_mingle:
        options.append("--without-mingle")
    if args.without_gossip:
        options.append("--without-gossip")

    if config.has_option("celery", "pool"):
        pool = config.get("celery", "pool")
        options.extend(["--pool", pool])
        # Celery pools of type eventlet and gevent use greenlets, which
        # requires monkey patching the app:
        # https://eventlet.readthedocs.io/en/latest/patching.html#monkeypatching-the-standard-library
        # Otherwise task instances hang on the workers and are never
        # executed.
        maybe_patch_concurrency(["-P", pool])

    worker_pid_file_path, stdout, stderr, log_file = setup_locations(
        process=WORKER_PROCESS_NAME,
        stdout=args.stdout,
        stderr=args.stderr,
        log=args.log_file,
        pid=args.pid,
    )

    def run_celery_worker():
        with _serve_logs(skip_serve_logs), _run_stale_bundle_cleanup():
            celery_app.worker_main(options)

    if args.umask:
        umask = args.umask
    else:
        umask = config.get("celery", "worker_umask", fallback=settings.DAEMON_UMASK)

    _run_command_with_daemon_option(
        args=args,
        process_name=WORKER_PROCESS_NAME,
        callback=run_celery_worker,
        should_setup_logging=True,
        umask=umask,
        pid_file=worker_pid_file_path,
    )


@cli_utils.action_cli(check_db=False)
@_providers_configuration_loaded
def stop_worker(args):
    """Send SIGTERM to Celery worker."""
    # Read PID from file
    if args.pid:
        pid_file_path = args.pid
    else:
        pid_file_path, _, _, _ = setup_locations(process=WORKER_PROCESS_NAME)
    pid = read_pid_from_pidfile(pid_file_path)

    # Send SIGTERM
    if pid:
        worker_process = psutil.Process(pid)
        worker_process.terminate()

    # Remove pid file
    remove_existing_pidfile(pid_file_path)


@_providers_configuration_loaded
def _check_if_active_celery_worker(hostname: str):
    """Check if celery worker is active before executing dependent cli commands."""
    # This needs to be imported locally to not trigger Providers Manager initialization
    from airflow.providers.celery.executors.celery_executor import app as celery_app

    inspect = celery_app.control.inspect()
    active_workers = inspect.active_queues()
    if not active_workers:
        raise SystemExit("Error: No active Celery workers found!")
    if hostname not in active_workers:
        raise SystemExit(f"Error: {hostname} is unknown!")


@cli_utils.action_cli(check_db=False)
@_providers_configuration_loaded
def list_workers(args):
    """List all active celery workers."""
    workers = []
    # This needs to be imported locally to not trigger Providers Manager initialization
    from airflow.providers.celery.executors.celery_executor import app as celery_app

    inspect = celery_app.control.inspect()
    active_workers = inspect.active_queues()
    if active_workers:
        workers = [
            {
                "worker_name": worker,
                "queues": [queue["name"] for queue in active_workers[worker] if "name" in queue],
            }
            for worker in active_workers
        ]
    AirflowConsole().print_as(data=workers, output=args.output)


@cli_utils.action_cli(check_db=False)
@_providers_configuration_loaded
def shutdown_worker(args):
    """Request graceful shutdown of a celery worker."""
    _check_if_active_celery_worker(hostname=args.celery_hostname)
    # This needs to be imported locally to not trigger Providers Manager initialization
    from airflow.providers.celery.executors.celery_executor import app as celery_app

    celery_app.control.shutdown(destination=[args.celery_hostname])


@cli_utils.action_cli(check_db=False)
@_providers_configuration_loaded
def shutdown_all_workers(args):
    """Request graceful shutdown all celery workers."""
    if not (
        args.yes
        or input(
            "This will shutdown all active celery workers connected to the celery broker, this cannot be undone! Proceed? (y/n)"
        ).upper()
        == "Y"
    ):
        raise SystemExit("Cancelled")
    # This needs to be imported locally to not trigger Providers Manager initialization
    from airflow.providers.celery.executors.celery_executor import app as celery_app

    celery_app.control.broadcast("shutdown")


@cli_utils.action_cli(check_db=False)
@_providers_configuration_loaded
def add_queue(args):
    """Subscribe a Celery worker to specified queues."""
    _check_if_active_celery_worker(hostname=args.celery_hostname)
    # This needs to be imported locally to not trigger Providers Manager initialization
    from airflow.providers.celery.executors.celery_executor import app as celery_app

    queues = args.queues.split(",")
    for queue in queues:
        celery_app.control.add_consumer(queue, destination=[args.celery_hostname])


@cli_utils.action_cli(check_db=False)
@_providers_configuration_loaded
def remove_queue(args):
    """Unsubscribe a Celery worker from specified queues."""
    _check_if_active_celery_worker(hostname=args.celery_hostname)
    # This needs to be imported locally to not trigger Providers Manager initialization
    from airflow.providers.celery.executors.celery_executor import app as celery_app

    queues = args.queues.split(",")
    for queue in queues:
        celery_app.control.cancel_consumer(queue, destination=[args.celery_hostname])


@cli_utils.action_cli(check_db=False)
@_providers_configuration_loaded
def remove_all_queues(args):
    """Unsubscribe a Celery worker from all its active queues."""
    _check_if_active_celery_worker(hostname=args.celery_hostname)
    # This needs to be imported locally to not trigger Providers Manager initialization
    from airflow.providers.celery.executors.celery_executor import app as celery_app

    inspect = celery_app.control.inspect()
    active_workers = inspect.active_queues()

    if not active_workers or args.celery_hostname not in active_workers:
        print(f"No active queues found for worker: {args.celery_hostname}")
        return

    worker_queues = active_workers[args.celery_hostname]
    queue_names = [queue["name"] for queue in worker_queues if "name" in queue]

    if not queue_names:
        print(f"No queues to remove for worker: {args.celery_hostname}")
        return

    print(
        f"Removing {len(queue_names)} queue(s) from worker {args.celery_hostname}: {', '.join(queue_names)}"
    )

    for queue_name in queue_names:
        celery_app.control.cancel_consumer(queue_name, destination=[args.celery_hostname])

    print(f"Successfully removed all queues from worker: {args.celery_hostname}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""CLI commands for Celery executor."""

from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.cli.cli_config import (
    ARG_DAEMON,
    ARG_LOG_FILE,
    ARG_PID,
    ARG_SKIP_SERVE_LOGS,
    ARG_STDERR,
    ARG_STDOUT,
    ARG_VERBOSE,
    ActionCommand,
    Arg,
    GroupCommand,
    lazy_load_command,
)
from airflow.configuration import conf

if TYPE_CHECKING:
    import argparse

# flower cli args
ARG_BROKER_API = Arg(("-a", "--broker-api"), help="Broker API")
ARG_FLOWER_HOSTNAME = Arg(
    ("-H", "--hostname"),
    default=conf.get("celery", "FLOWER_HOST"),
    help="Set the hostname on which to run the server",
)
ARG_FLOWER_PORT = Arg(
    ("-p", "--port"),
    default=conf.getint("celery", "FLOWER_PORT"),
    type=int,
    help="The port on which to run the server",
)
ARG_FLOWER_CONF = Arg(("-c", "--flower-conf"), help="Configuration file for flower")
ARG_FLOWER_URL_PREFIX = Arg(
    ("-u", "--url-prefix"),
    default=conf.get("celery", "FLOWER_URL_PREFIX"),
    help="URL prefix for Flower",
)
ARG_FLOWER_BASIC_AUTH = Arg(
    ("-A", "--basic-auth"),
    default=conf.get("celery", "FLOWER_BASIC_AUTH"),
    help=(
        "Securing Flower with Basic Authentication. "
        "Accepts user:password pairs separated by a comma. "
        "Example: flower_basic_auth = user1:password1,user2:password2"
    ),
)

# worker cli args
ARG_AUTOSCALE = Arg(("-a", "--autoscale"), help="Minimum and Maximum number of worker to autoscale")
ARG_QUEUES = Arg(
    ("-q", "--queues"),
    help="Comma delimited list of queues to serve",
    default=conf.get("operators", "DEFAULT_QUEUE"),
)
ARG_CONCURRENCY = Arg(
    ("-c", "--concurrency"),
    type=int,
    help="The number of worker processes",
    default=conf.getint("celery", "worker_concurrency"),
)
ARG_CELERY_HOSTNAME = Arg(
    ("-H", "--celery-hostname"),
    help="Set the hostname of celery worker if you have multiple workers on a single machine",
)
ARG_UMASK = Arg(
    ("-u", "--umask"),
    help="Set the umask of celery worker in daemon mode",
)

ARG_WITHOUT_MINGLE = Arg(
    ("--without-mingle",),
    default=False,
    help="Don't synchronize with other workers at start-up",
    action="store_true",
)
ARG_WITHOUT_GOSSIP = Arg(
    ("--without-gossip",),
    default=False,
    help="Don't subscribe to other workers events",
    action="store_true",
)
ARG_TEAM = Arg(
    ("-t", "--team"),
    help="Team name for team-specific multi-team configuration (requires Airflow 3.2+)",
)
ARG_OUTPUT = Arg(
    (
        "-o",
        "--output",
    ),
    help="Output format. Allowed values: json, yaml, plain, table (default: table)",
    metavar="(table, json, yaml, plain)",
    choices=("table", "json", "yaml", "plain"),
    default="table",
)
ARG_FULL_CELERY_HOSTNAME = Arg(
    ("-H", "--celery-hostname"),
    required=True,
    help="Specify the full celery hostname. example: celery@hostname",
)
ARG_REQUIRED_QUEUES = Arg(
    ("-q", "--queues"),
    help="Comma delimited list of queues to serve",
    required=True,
)
ARG_YES = Arg(
    ("-y", "--yes"),
    help="Do not prompt to confirm. Use with care!",
    action="store_true",
    default=False,
)

CELERY_CLI_COMMAND_PATH = "airflow.providers.celery.cli.celery_command"

CELERY_COMMANDS = (
    ActionCommand(
        name="worker",
        help="Start a Celery worker node",
        func=lazy_load_command(f"{CELERY_CLI_COMMAND_PATH}.worker"),
        args=(
            ARG_QUEUES,
            ARG_CONCURRENCY,
            ARG_CELERY_HOSTNAME,
            ARG_TEAM,
            ARG_PID,
            ARG_DAEMON,
            ARG_UMASK,
            ARG_STDOUT,
            ARG_STDERR,
            ARG_LOG_FILE,
            ARG_AUTOSCALE,
            ARG_SKIP_SERVE_LOGS,
            ARG_WITHOUT_MINGLE,
            ARG_WITHOUT_GOSSIP,
            ARG_VERBOSE,
        ),
    ),
    ActionCommand(
        name="flower",
        help="Start a Celery Flower",
        func=lazy_load_command(f"{CELERY_CLI_COMMAND_PATH}.flower"),
        args=(
            ARG_FLOWER_HOSTNAME,
            ARG_FLOWER_PORT,
            ARG_FLOWER_CONF,
            ARG_FLOWER_URL_PREFIX,
            ARG_FLOWER_BASIC_AUTH,
            ARG_BROKER_API,
            ARG_PID,
            ARG_DAEMON,
            ARG_STDOUT,
            ARG_STDERR,
            ARG_LOG_FILE,
            ARG_VERBOSE,
        ),
    ),
    ActionCommand(
        name="stop",
        help="Stop the Celery worker gracefully",
        func=lazy_load_command(f"{CELERY_CLI_COMMAND_PATH}.stop_worker"),
        args=(ARG_PID, ARG_VERBOSE),
    ),
    ActionCommand(
        name="list-workers",
        help="List active celery workers",
        func=lazy_load_command(f"{CELERY_CLI_COMMAND_PATH}.list_workers"),
        args=(ARG_OUTPUT,),
    ),
    ActionCommand(
        name="shutdown-worker",
        help="Request graceful shutdown of celery workers",
        func=lazy_load_command(f"{CELERY_CLI_COMMAND_PATH}.shutdown_worker"),
        args=(ARG_FULL_CELERY_HOSTNAME,),
    ),
    ActionCommand(
        name="shutdown-all-workers",
        help="Request graceful shutdown of all active celery workers",
        func=lazy_load_command(f"{CELERY_CLI_COMMAND_PATH}.shutdown_all_workers"),
        args=(ARG_YES,),
    ),
    ActionCommand(
        name="add-queue",
        help="Subscribe Celery worker to specified queues",
        func=lazy_load_command(f"{CELERY_CLI_COMMAND_PATH}.add_queue"),
        args=(
            ARG_REQUIRED_QUEUES,
            ARG_FULL_CELERY_HOSTNAME,
        ),
    ),
    ActionCommand(
        name="remove-queue",
        help="Unsubscribe Celery worker from specified queues",
        func=lazy_load_command(f"{CELERY_CLI_COMMAND_PATH}.remove_queue"),
        args=(
            ARG_REQUIRED_QUEUES,
            ARG_FULL_CELERY_HOSTNAME,
        ),
    ),
    ActionCommand(
        name="remove-all-queues",
        help="Unsubscribe Celery worker from all its active queues",
        func=lazy_load_command(f"{CELERY_CLI_COMMAND_PATH}.remove_all_queues"),
        args=(ARG_FULL_CELERY_HOSTNAME,),
    ),
)

CELERY_CLI_COMMANDS = [
    GroupCommand(
        name="celery",
        help="Celery components",
        description=(
            "Start celery components. Works only when using CeleryExecutor. For more information, "
            "see https://airflow.apache.org/docs/apache-airflow/stable/executor/celery.html"
        ),
        subcommands=CELERY_COMMANDS,
    ),
]


def get_celery_cli_commands():
    """Return CLI commands for Celery executor."""
    return CELERY_CLI_COMMANDS


def get_parser() -> argparse.ArgumentParser:
    """
    Generate documentation; used by Sphinx.

    :meta private:
    """
    from airflow.cli.cli_parser import AirflowHelpFormatter, DefaultHelpParser, _add_command

    parser = DefaultHelpParser(prog="airflow", formatter_class=AirflowHelpFormatter)
    subparsers = parser.add_subparsers(dest="subcommand", metavar="GROUP_OR_COMMAND")
    for group_command in get_celery_cli_commands():
        _add_command(subparsers, group_command)
    return parser
